from typing import Any, Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import (
    LanguageModelOutput,
    create_attention_mask,
    scaled_dot_product_attention,
)
from .config import ModelConfig, TextConfig


def _compute_default_rope_parameters(
    config: Optional[TextConfig] = None,
    **rope_kwargs,
) -> tuple[mx.array, float]:

    if config is not None and len(rope_kwargs) > 0:
        raise ValueError(
            "Unexpected arguments: `**rope_kwargs` and `config` are mutually exclusive in "
            f"`_compute_default_rope_parameters`, got `rope_kwargs`={rope_kwargs} and `config`={config}"
        )
    if len(rope_kwargs) > 0:
        base = rope_kwargs["base"]
        dim = rope_kwargs["dim"]
    elif config is not None:
        base = config.rope_theta
        partial_rotary_factor = config.partial_rotary_factor
        head_dim = config.head_dim
        dim = int(head_dim * partial_rotary_factor)

    attention_factor = 1.0

    inv_freq = 1.0 / (
        base ** (mx.arange(0, dim, 2, dtype=mx.int64).astype(mx.float32) / dim)
    )
    return inv_freq, attention_factor


class GlmOcrRotaryEmbedding(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()

        self.rope_type = config.rope_parameters.get("rope_type", "default")
        self.max_seq_len_cached = config.max_position_embeddings
        self.original_max_seq_len = config.max_position_embeddings

        self.config = config
        self.mrope_section = config.rope_parameters.get("mrope_section", [16, 24, 24])

        self.rope_init_fn = _compute_default_rope_parameters

        inv_freq, self.attention_scaling = self.rope_init_fn(self.config)
        self._inv_freq = mx.array(inv_freq, dtype=mx.float32)
        self._original_inv_freq = mx.array(inv_freq, dtype=mx.float32)

    def apply_mrope(self, freqs, mrope_section):
        """Apply M-RoPE by selecting different dimensions for T, H, W."""
        split_indices = np.cumsum(mrope_section)[:-1].tolist()
        chunks = mx.split(freqs, split_indices, axis=-1)
        result = mx.concatenate(
            [chunk[i % 3] for i, chunk in enumerate(chunks)], axis=-1
        )
        return result

    def __call__(self, x, position_ids):
        inv_freq_expanded = self._inv_freq[None, None, :, None].astype(mx.float32)
        inv_freq_expanded = mx.broadcast_to(
            inv_freq_expanded, (3, position_ids.shape[1], self._inv_freq.shape[0], 1)
        )
        position_ids_expanded = position_ids[:, :, None, :].astype(mx.float32)

        freqs = (
            inv_freq_expanded.astype(mx.float32)
            @ position_ids_expanded.astype(mx.float32)
        ).transpose(0, 1, 3, 2)

        freqs = self.apply_mrope(freqs, self.mrope_section)

        emb = mx.concatenate((freqs, freqs), axis=-1)
        cos = mx.cos(emb) * self.attention_scaling
        sin = mx.sin(emb) * self.attention_scaling

        return cos.astype(x.dtype), sin.astype(x.dtype)


def rotate_half_llm(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., 0::2]
    x2 = x[..., 1::2]
    return mx.flatten(mx.stack([-x2, x1], axis=-1), start_axis=-2, end_axis=-1)


def repeat_interleave(x, repeats, axis=-1):
    """
    Repeat elements of an array along an axis, interleaving the repeated values.
    Like torch.repeat_interleave: [a,b,c] with repeats=2 -> [a,a,b,b,c,c]
    """
    shape = list(x.shape)
    x = mx.expand_dims(x, axis=axis + 1 if axis >= 0 else axis)
    tile_shape = [1] * len(x.shape)
    tile_shape[axis + 1 if axis >= 0 else axis] = repeats
    x = mx.tile(x, tile_shape)
    new_shape = shape.copy()
    new_shape[axis] = shape[axis] * repeats
    return x.reshape(new_shape)


def apply_rotary_pos_emb(q, k, cos, sin):
    """
    Applies Rotary Position Embedding to the query and key tensors.
    Matches PyTorch's GLM-OCR implementation exactly.

    Args:
        q: Query tensor of shape (batch, n_heads, seq_len, head_dim)
        k: Key tensor of shape (batch, n_kv_heads, seq_len, head_dim)
        cos: Cosine tensor of shape (batch, seq_len, head_dim)
        sin: Sine tensor of shape (batch, seq_len, head_dim)
    """
    cos = cos[:, None, :, :]
    sin = sin[:, None, :, :]

    cos = repeat_interleave(cos[..., : cos.shape[-1] // 2], repeats=2, axis=-1)
    sin = repeat_interleave(sin[..., : sin.shape[-1] // 2], repeats=2, axis=-1)

    rotary_dim = cos.shape[-1]
    q_rot = q[..., :rotary_dim]
    q_pass = q[..., rotary_dim:]

    k_rot = k[..., :rotary_dim]
    k_pass = k[..., rotary_dim:]

    q_embed = (q_rot * cos) + (rotate_half_llm(q_rot) * sin)
    k_embed = (k_rot * cos) + (rotate_half_llm(k_rot) * sin)

    q_embed = mx.concatenate([q_embed, q_pass], axis=-1)
    k_embed = mx.concatenate([k_embed, k_pass], axis=-1)

    return q_embed, k_embed


class GlmOcrAttention(nn.Module):
    def __init__(self, args: TextConfig):
        super().__init__()

        dim = args.hidden_size
        self.n_heads = n_heads = args.num_attention_heads
        self.n_kv_heads = n_kv_heads = args.num_key_value_heads
        self.head_dim = args.head_dim
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(dim, n_heads * self.head_dim, bias=args.attention_bias)
        self.k_proj = nn.Linear(
            dim, n_kv_heads * self.head_dim, bias=args.attention_bias
        )
        self.v_proj = nn.Linear(
            dim, n_kv_heads * self.head_dim, bias=args.attention_bias
        )
        self.o_proj = nn.Linear(n_heads * self.head_dim, dim, bias=False)

        self.rope_parameters = args.rope_parameters

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Any] = None,
        position_embeddings: Optional[mx.array] = None,
    ) -> mx.array:
        B, L, _ = x.shape

        queries = self.q_proj(x)
        keys = self.k_proj(x)
        values = self.v_proj(x)

        queries = queries.reshape(B, L, self.n_heads, -1)
        keys = keys.reshape(B, L, self.n_kv_heads, -1)

        queries = queries.transpose(0, 2, 1, 3)
        keys = keys.transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        cos, sin = position_embeddings

        queries, keys = apply_rotary_pos_emb(queries, keys, cos, sin)

        if cache is not None:
            keys, values = cache.update_and_fetch(keys, values)

        if mask is not None and isinstance(mask, mx.array):
            mask = mask[..., : keys.shape[-2]]

        output = scaled_dot_product_attention(
            queries, keys, values, cache=cache, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output)


class GlmOcrMLP(nn.Module):
    def __init__(
        self, config: TextConfig, hidden_size: int = None, intermediate_size: int = None
    ):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size if hidden_size is None else hidden_size
        self.intermediate_size = (
            config.intermediate_size if intermediate_size is None else intermediate_size
        )

        self.gate_up_proj = nn.Linear(
            self.hidden_size, self.intermediate_size * 2, bias=False
        )
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)

    def __call__(self, x):
        x = self.gate_up_proj(x)
        gate, x = mx.split(x, 2, axis=-1)
        return self.down_proj(nn.silu(gate) * x)


class GlmOcrDecoderLayer(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.self_attn = GlmOcrAttention(config)
        self.mlp = GlmOcrMLP(config)

        self.input_layernorm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = nn.RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.post_self_attn_layernorm = nn.RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.post_mlp_layernorm = nn.RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Any] = None,
        position_embeddings: Optional[mx.array] = None,
    ) -> mx.array:
        r = x

        x = self.self_attn(self.input_layernorm(x), mask, cache, position_embeddings)

        x = self.post_self_attn_layernorm(x)
        x = r + x

        r = x
        x = self.post_attention_layernorm(x)
        x = self.mlp(x)
        x = self.post_mlp_layernorm(x)
        x = r + x
        return x


class GlmOcrTextModel(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.vocab_size = config.vocab_size
        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size)
        self.layers = [
            GlmOcrDecoderLayer(config) for _ in range(config.num_hidden_layers)
        ]
        self.start_idx = 0
        self.end_idx = len(self.layers)
        self.num_layers = self.end_idx

        self.norm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

        self.rotary_emb = GlmOcrRotaryEmbedding(config)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        cache: Optional[Any] = None,
        mask: Optional[mx.array] = None,
        position_ids: Optional[mx.array] = None,
    ) -> mx.array:

        if inputs_embeds is None:
            h = self.embed_tokens(inputs)
        else:
            h = inputs_embeds.astype(self.norm.weight.dtype)

        if position_ids is None:
            position_ids = mx.arange(cache[0].offset, cache[0].offset + h.shape[-2])
            position_ids = mx.expand_dims(position_ids, axis=0)
            position_ids = mx.tile(position_ids, (3, 1, 1))

        position_embeddings = self.rotary_emb(h, position_ids)

        if mask is None:
            mask = create_attention_mask(h, cache)

        if cache is None:
            cache = [None] * self.num_layers

        for i in range(self.num_layers):
            h = self.layers[self.start_idx + i](h, mask, cache[i], position_embeddings)

        return self.norm(h)


class LanguageModel(nn.Module):
    def __init__(self, args: TextConfig, config: ModelConfig = None):
        super().__init__()
        self.args = args
        self.config = config
        self.model_type = args.model_type
        self.model = GlmOcrTextModel(args)
        self.lm_head = nn.Linear(args.hidden_size, args.vocab_size, bias=False)
        self._rope_deltas = None
        self._position_ids = None

    def get_rope_index(
        self,
        input_ids: mx.array,
        image_grid_thw: Optional[mx.array] = None,
        video_grid_thw: Optional[mx.array] = None,
        attention_mask: Optional[mx.array] = None,
    ):
        batch_size, seq_length = input_ids.shape
        position_ids = mx.arange(seq_length, dtype=mx.int32)
        position_ids = mx.broadcast_to(position_ids[None, :], (batch_size, seq_length))
        spatial_merge_size = self.config.vision_config.spatial_merge_size
        image_token_id = self.config.image_token_id
        video_token_id = self.config.video_token_id
        image_start_token_id = self.config.image_start_token_id
        mrope_position_deltas = []
        if input_ids is not None and (
            image_grid_thw is not None or video_grid_thw is not None
        ):
            total_input_ids = input_ids
            if (
                attention_mask is None
                or attention_mask.shape[-1] != input_ids.shape[-1]
            ):
                attention_mask = mx.ones_like(input_ids)
            position_ids = mx.ones(
                (3, input_ids.shape[0], input_ids.shape[1]), dtype=input_ids.dtype
            )
            image_index, video_index = 0, 0
            for i, input_ids in enumerate(total_input_ids):
                input_ids = mx.where(
                    attention_mask[i] == 1, input_ids, mx.zeros_like(input_ids)
                )
                image_nums, video_nums = 0, 0
                vision_start_indices = mx.sum(
                    mx.where(
                        input_ids == image_start_token_id,
                        mx.arange(input_ids.shape[0]),
                        mx.zeros_like(input_ids),
                    )
                )
                vision_tokens = input_ids[vision_start_indices + 1]
                image_nums = (vision_tokens == image_token_id).sum().item()
                video_nums = (vision_tokens == video_token_id).sum().item()
                input_tokens = input_ids.tolist()
                llm_pos_ids_list: list = []
                st = 0
                remain_images, remain_videos = image_nums, video_nums
                for _ in range(image_nums + video_nums):
                    if image_token_id in input_tokens and remain_images > 0:
                        ed_image = input_tokens.index(image_token_id, st)
                    else:
                        ed_image = len(input_tokens) + 1
                    if video_token_id in input_tokens and remain_videos > 0:
                        ed_video = input_tokens.index(video_token_id, st)
                    else:
                        ed_video = len(input_tokens) + 1
                    if ed_image < ed_video:
                        t, h, w = (
                            image_grid_thw[image_index][0],
                            image_grid_thw[image_index][1],
                            image_grid_thw[image_index][2],
                        )
                        image_index += 1
                        remain_images -= 1
                        ed = ed_image
                    else:
                        t, h, w = (
                            video_grid_thw[video_index][0],
                            video_grid_thw[video_index][1],
                            video_grid_thw[video_index][2],
                        )
                        video_index += 1
                        remain_videos -= 1
                        ed = ed_video
                    llm_grid_t, llm_grid_h, llm_grid_w = (
                        t.item(),
                        h.item() // spatial_merge_size,
                        w.item() // spatial_merge_size,
                    )
                    text_len = ed - st
                    st_idx = (
                        llm_pos_ids_list[-1].max() + 1
                        if len(llm_pos_ids_list) > 0
                        else 0
                    )
                    index = mx.arange(text_len).reshape(1, text_len)
                    index = mx.broadcast_to(index, (3, text_len))
                    index = index + st_idx
                    llm_pos_ids_list.append(index)
                    t_index = mx.arange(llm_grid_t).reshape(llm_grid_t, 1)
                    t_index = mx.broadcast_to(
                        t_index, (llm_grid_t, llm_grid_h * llm_grid_w)
                    )
                    t_index = t_index.flatten()

                    h_index = mx.arange(llm_grid_h).reshape(1, llm_grid_h, 1)
                    h_index = mx.broadcast_to(
                        h_index, (llm_grid_t, llm_grid_h, llm_grid_w)
                    )
                    h_index = h_index.flatten()

                    w_index = mx.arange(llm_grid_w).reshape(1, 1, llm_grid_w)
                    w_index = mx.broadcast_to(
                        w_index, (llm_grid_t, llm_grid_h, llm_grid_w)
                    )
                    w_index = w_index.flatten()

                    llm_pos_ids_list.append(
                        mx.stack([t_index, h_index, w_index]) + text_len + st_idx
                    )
                    st = ed + llm_grid_t * llm_grid_h * llm_grid_w
                if st < len(input_tokens):
                    st_idx = (
                        llm_pos_ids_list[-1].max() + 1
                        if len(llm_pos_ids_list) > 0
                        else 0
                    )
                    text_len = len(input_tokens) - st

                    t_index = mx.arange(text_len).reshape(1, text_len)
                    t_index = mx.broadcast_to(t_index, (3, text_len))

                    llm_pos_ids_list.append(t_index + st_idx)

                llm_positions = mx.concatenate(llm_pos_ids_list, axis=1).reshape(3, -1)
                mask = mx.array(attention_mask[i] == 1)
                expanded_mask = mx.expand_dims(mask, axis=0)
                expanded_mask = mx.broadcast_to(expanded_mask, (3, 1, mask.shape[0]))
                expanded_positions = mx.expand_dims(llm_positions, axis=1)
                new_positions = mx.where(
                    expanded_mask, expanded_positions, position_ids[:, i : i + 1, :]
                )
                updated_position_ids = mx.concatenate(
                    [
                        position_ids[:, :i, :],
                        new_positions,
                        position_ids[:, i + 1 :, :],
                    ],
                    axis=1,
                )
                position_ids = updated_position_ids
                mrope_position_deltas.append(
                    llm_positions.max() + 1 - len(total_input_ids[i])
                )
            mrope_position_deltas = mx.array(mrope_position_deltas)[0]
            return position_ids, mrope_position_deltas
        else:
            if attention_mask is not None:
                position_ids = mx.cumsum(attention_mask.astype(mx.int64), axis=-1) - 1
                position_ids = mx.where(
                    attention_mask == 0, mx.ones_like(position_ids), position_ids
                )
                position_ids = mx.expand_dims(position_ids[0], axis=0)
                position_ids = mx.tile(position_ids, (3, 1, 1))
                max_position_ids = position_ids.max(0, keepdims=False)[0].max(
                    -1, keepdims=True
                )[0]
                mrope_position_deltas = max_position_ids + 1 - attention_mask.shape[-1]
            else:
                position_ids = mx.arange(input_ids.shape[1]).reshape(1, -1)
                position_ids = mx.broadcast_to(
                    position_ids, (3, input_ids.shape[0], input_ids.shape[1])
                )
                mrope_position_deltas = mx.zeros(
                    [input_ids.shape[0], 1],
                    dtype=input_ids.dtype,
                )
            return position_ids, mrope_position_deltas

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):

        position_ids = kwargs.pop("position_ids", None)
        pixel_values = kwargs.pop("pixel_values", None)
        image_grid_thw = kwargs.pop("image_grid_thw", None)
        video_grid_thw = kwargs.pop("video_grid_thw", None)
        if pixel_values is not None:
            self._rope_deltas = None

        cache_offset = 0
        if cache and cache[0] is not None:
            offset = cache[0].offset
            if isinstance(offset, int):
                cache_offset = offset
            elif isinstance(offset, mx.array):
                cache_offset = (offset if offset.ndim == 0 else offset[0]).item()
            else:
                raise ValueError(f"Unexpected cache offset type: {type(offset)}")

        # Check if mask shape matches input shape (for chunked prefill compatibility)
        rope_mask = mask
        if mask is not None and mask.shape[-1] != inputs.shape[-1]:
            rope_mask = None

        if position_ids is None and (rope_mask is None or rope_mask.ndim == 2):
            # Calculate RoPE index once per generation in the pre-fill stage only
            if (
                (cache is not None and cache[0] is not None and (cache_offset == 0))
                or self._rope_deltas is None
                or cache is None
            ):
                # Use cached position_ids if available (pre-computed in get_input_embeddings)
                if self._position_ids is not None:
                    seq_length = inputs.shape[1]
                    position_ids = self._position_ids[
                        :, :, cache_offset : cache_offset + seq_length
                    ]
                else:
                    position_ids, rope_deltas = self.get_rope_index(
                        inputs, image_grid_thw, video_grid_thw, rope_mask
                    )
                    self._rope_deltas = rope_deltas
                    self._position_ids = position_ids
            else:
                # Use the prev pre-calculated rope-deltas to get the correct position ids
                batch_size, seq_length = inputs.shape
                delta = mx.array(
                    cache_offset + self._rope_deltas if cache is not None else 0
                )
                position_ids = mx.arange(seq_length).reshape(1, -1)
                position_ids = mx.broadcast_to(position_ids, (batch_size, seq_length))

                if cache_offset is not None:
                    if delta.ndim == 0:
                        delta = mx.expand_dims(delta, axis=0)

                    if delta.shape[0] < batch_size:
                        delta = mx.tile(delta, (batch_size, 1))
                    else:
                        delta = delta[:batch_size]

                position_ids = mx.add(position_ids, delta)[None, ...]
                position_ids = mx.broadcast_to(
                    position_ids, (3, batch_size, seq_length)
                )

        out = self.model(
            inputs,
            cache=cache,
            inputs_embeds=inputs_embeds,
            position_ids=position_ids,
            mask=mask,
        )

        out = self.lm_head(out)
        return LanguageModelOutput(logits=out)

    def sanitize(self, weights):
        return weights

    @property
    def layers(self):
        return self.model.layers

    @property
    def n_kv_heads(self):
        return self.args.num_key_value_heads



from .config import ModelConfig, TextConfig, VisionConfig
from .ernie4_5_moe_vl import Model, VariableResolutionResamplerModel
from .language import LanguageModel
from .processor import Ernie4_5_VLProcessor, Ernie4_5_VLTokenizer, ImageProcessor
from .vision import VisionModel



from mlx_vlm import generate
from mlx_vlm.prompt_utils import apply_chat_template


def inference(
    model,
    processor,
    question,
    image,
    max_tokens=3000,
    temperature=0.0,
    resize_shape=None,
    verbose=False,
):
    """Run inference on a single question."""
    if image is None:
        num_images = 0
    elif isinstance(image, list):
        num_images = len(image)
    else:
        num_images = 1

    prompt = apply_chat_template(
        processor, model.config, question, num_images=num_images
    )

    response = generate(
        model,
        processor,
        prompt,
        image=image,
        max_tokens=max_tokens,
        temperature=temperature,
        resize_shape=resize_shape,
        verbose=verbose,
    )
    return response.text



"""
From https://github.com/deepseek-ai/DeepSeek-VL2
"""

from dataclasses import dataclass
from typing import Dict, List, Literal, Optional, Tuple

import mlx.core as mx
import numpy as np
from PIL import Image, ImageOps
from transformers import LlamaTokenizerFast
from transformers.processing_utils import ProcessorMixin


def find_closest_aspect_ratio(aspect_ratio, target_ratios, width, height, image_size):
    best_ratio_diff = float("inf")
    best_ratio = (1, 1)
    area = width * height
    for ratio in target_ratios:
        target_aspect_ratio = ratio[0] / ratio[1]
        ratio_diff = abs(aspect_ratio - target_aspect_ratio)
        if ratio_diff < best_ratio_diff:
            best_ratio_diff = ratio_diff
            best_ratio = ratio
        elif ratio_diff == best_ratio_diff:
            if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                best_ratio = ratio
    # print(f'width: {width}, height: {height}, best_ratio: {best_ratio}')
    return best_ratio


def dynamic_preprocess(
    image, min_num=2, max_num=9, image_size=640, use_thumbnail=False
):
    orig_width, orig_height = image.size
    aspect_ratio = orig_width / orig_height

    # calculate the existing image aspect ratio
    target_ratios = set(
        (i, j)
        for n in range(min_num, max_num + 1)
        for i in range(1, n + 1)
        for j in range(1, n + 1)
        if i * j <= max_num and i * j >= min_num
    )
    # print(target_ratios)
    target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

    # find the closest aspect ratio to the target
    target_aspect_ratio = find_closest_aspect_ratio(
        aspect_ratio, target_ratios, orig_width, orig_height, image_size
    )

    # print(target_aspect_ratio)
    # calculate the target width and height
    target_width = image_size * target_aspect_ratio[0]
    target_height = image_size * target_aspect_ratio[1]
    blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

    # resize the image
    resized_img = image.resize((target_width, target_height))
    processed_images = []
    for i in range(blocks):
        box = (
            (i % (target_width // image_size)) * image_size,
            (i // (target_width // image_size)) * image_size,
            ((i % (target_width // image_size)) + 1) * image_size,
            ((i // (target_width // image_size)) + 1) * image_size,
        )
        # split the image
        split_img = resized_img.crop(box)
        processed_images.append(split_img)
    assert len(processed_images) == blocks
    if use_thumbnail and len(processed_images) != 1:
        thumbnail_img = image.resize((image_size, image_size))
        processed_images.append(thumbnail_img)
    return processed_images, target_aspect_ratio


class DictOutput(object):
    def keys(self):
        return self.__dict__.keys()

    def __getitem__(self, item):
        if isinstance(item, int):
            return list(self.__dict__.values())[item]
        if item not in self.__dict__:
            raise KeyError(item)
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value


@dataclass
class VLChatProcessorOutput(DictOutput):
    sft_format: str
    input_ids: mx.array
    target_ids: mx.array
    images: mx.array
    images_seq_mask: mx.array
    images_spatial_crop: mx.array
    num_image_tokens: List[int]

    def __len__(self):
        return len(self.input_ids)


@dataclass
class BatchCollateOutput(DictOutput):
    sft_format: List[str]
    input_ids: mx.array
    labels: mx.array
    images: mx.array
    attention_mask: mx.array
    images_seq_mask: mx.array
    images_spatial_crop: mx.array
    seq_lens: List[int]


class ImageTransform:
    def __init__(
        self,
        mean: Optional[Tuple[float, float, float]] = (0.5, 0.5, 0.5),
        std: Optional[Tuple[float, float, float]] = (0.5, 0.5, 0.5),
        normalize: bool = True,
    ):
        self.mean = mean
        self.std = std
        self.normalize = normalize

    def __call__(self, pil_img: Image.Image):
        # Convert PIL image to numpy array and normalize

        img = mx.array(np.array(pil_img)) / 255.0

        # Transpose from HWC to CHW format
        img = mx.transpose(img, [2, 0, 1])

        if self.normalize:
            mean = mx.array(self.mean).reshape(-1, 1, 1)
            std = mx.array(self.std).reshape(-1, 1, 1)
            img = (img - mean) / std

        return img


class DeepseekOCR2Processor(ProcessorMixin):
    tokenizer_class = ("LlamaTokenizer", "LlamaTokenizerFast")
    attributes = ["tokenizer"]

    def __init__(
        self,
        tokenizer: LlamaTokenizerFast,
        candidate_resolutions: Tuple[Tuple[int, int]],
        patch_size: int,
        downsample_ratio: int,
        image_mean: Tuple[float, float, float] = (0.5, 0.5, 0.5),
        image_std: Tuple[float, float, float] = (0.5, 0.5, 0.5),
        normalize: bool = True,
        image_token: str = "<image>",
        pad_token: str = "<｜▁pad▁｜>",
        add_special_token: bool = False,
        sft_format: str = "deepseek",
        mask_prompt: bool = True,
        ignore_id: int = -100,
        **kwargs,
    ):
        self.candidate_resolutions = candidate_resolutions
        self.image_size = candidate_resolutions[0][0]
        self.patch_size = patch_size
        self.image_mean = image_mean
        self.image_std = image_std
        self.normalize = normalize
        self.downsample_ratio = downsample_ratio

        self.image_transform = ImageTransform(
            mean=image_mean, std=image_std, normalize=normalize
        )
        self.tokenizer = tokenizer
        self.tokenizer.padding_side = "left"

        # Add special tokens
        if tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": pad_token})
        print(
            f"Add pad token = ['{pad_token}'] to the tokenizer\n"
            f"{pad_token}:{tokenizer.encode(pad_token, add_special_tokens=False)[0]}"
        )

        image_token_id = self.tokenizer.vocab.get(image_token)
        if image_token_id is None:
            special_tokens = [image_token]
            special_tokens_dict = {"additional_special_tokens": special_tokens}
            self.tokenizer.add_special_tokens(special_tokens_dict)
        self.image_token_id = self.tokenizer.vocab.get(image_token)
        print(
            f"Add image token = ['{image_token}'] to the tokenizer\n"
            f"{image_token}:{tokenizer.encode(image_token, add_special_tokens=False)[0]}"
        )

        # Add grounding-related tokens
        special_tokens = ["<|ref|>", "<|/ref|>", "<|det|>", "<|/det|>", "<|grounding|>"]
        special_tokens_dict = {"additional_special_tokens": special_tokens}
        self.tokenizer.add_special_tokens(special_tokens_dict)
        print("Added grounding-related tokens")

        # Add chat tokens
        special_tokens = ["<|User|>", "<|Assistant|>"]
        special_tokens_dict = {"additional_special_tokens": special_tokens}
        self.tokenizer.add_special_tokens(special_tokens_dict)
        print("Added chat tokens")

        self.image_token = image_token
        self.pad_token = pad_token
        self.add_special_token = add_special_token
        self.sft_format = sft_format
        self.mask_prompt = mask_prompt
        self.ignore_id = ignore_id

        super().__init__(tokenizer, **kwargs)

        # Add chat template
        self.chat_template = kwargs.pop("chat_template", self.default_chat_template)

    @property
    def default_chat_template(self):
        return (
            "{% for message in messages %}"
            "{% if message['role'] == 'user' %}"
            "{% elif message['role'] == 'assistant' %}{% endif %}"
            "{{message['content']}} "
            "{% endfor %}"
            "{% if add_generation_prompt %}{% endif %}"
        )

    @property
    def bos_id(self):
        return self.tokenizer.bos_token_id

    @property
    def eos_id(self):
        return self.tokenizer.eos_token_id

    @property
    def pad_id(self):
        return self.tokenizer.pad_token_id

    def encode(self, text: str, bos: bool = True, eos: bool = False):
        t = self.tokenizer.encode(text, add_special_tokens=False)

        if bos:
            t = [self.bos_id] + t
        if eos:
            t = t + [self.eos_id]

        return t

    def decode(self, t: List[int], **kwargs) -> str:
        return self.tokenizer.decode(t, **kwargs)

    def process_one(
        self,
        prompt: str = None,
        images: List[Image.Image] = None,
        inference_mode: bool = True,
        base_size: int = 1024,
        image_size: int = 768,
        cropping: bool = True,
        min_patches: int = 1,
        max_patches: int = 6,
    ):

        sft_format = prompt
        (
            tokenized_str,
            images_list,
            images_seq_mask,
            images_spatial_crop,
            num_image_tokens,
        ) = self.tokenize_with_images(
            sft_format,
            images,
            base_size=base_size,
            image_size=image_size,
            cropping=cropping,
            min_patches=min_patches,
            max_patches=max_patches,
        )

        masked_tokenized_str = []
        for token_index in tokenized_str:
            if token_index != self.image_token_id:
                masked_tokenized_str.append(token_index)
            else:
                masked_tokenized_str.append(self.ignore_id)

        input_ids = mx.array(tokenized_str)
        target_ids = mx.array(masked_tokenized_str)
        images_seq_mask = mx.array(images_seq_mask)

        # Set ignored indices
        target_ids = mx.where(
            (input_ids < 0) | (input_ids == self.image_token_id),
            self.ignore_id,
            target_ids,
        )
        input_ids = mx.where(input_ids < 0, self.pad_id, input_ids)

        if inference_mode:
            input_ids = input_ids[:-1]
            target_ids = target_ids[:-1]
            images_seq_mask = images_seq_mask[:-1]

        return {
            "input_ids": input_ids[None, :],
            "attention_mask": input_ids != self.pad_id,
            "labels": target_ids,
            "images": images_list,
            "images_seq_mask": images_seq_mask[None, ...],
            "images_spatial_crop": images_spatial_crop,
            "num_image_tokens": num_image_tokens,
        }

    def pad_sequence(self, sequences, padding_value):
        # Get max length of sequences
        max_len = max(len(seq) for seq in sequences)

        # Pad each sequence to max length
        padded_seqs = []
        for seq in sequences:
            pad_length = max_len - len(seq)
            if pad_length > 0:
                padding = mx.full((pad_length,), padding_value)
                padded_seq = mx.concatenate([seq, padding])
            else:
                padded_seq = seq
            padded_seqs.append(padded_seq)

        return mx.stack(padded_seqs)

    def tokenize_with_images(
        self,
        conversation: str,
        images: List[Image.Image],
        base_size: int = 1024,
        image_size: int = 768,
        cropping: bool = True,
        min_patches: int = 1,
        max_patches: int = 6,
    ):
        """Tokenize text with <image> tags.

        For DeepSeek-OCR-2 with Qwen2 encoder:
        - Global view (1024x1024): 256 tokens from Qwen2 encoder
        - Local patches (768x768): 144 tokens each from Qwen2 encoder
        - Plus 1 view_separator token

        Dynamic resolution:
        - Total tokens = (num_patches * 144) + 256 + 1
        - Default: 0-6 patches at 768x768 + 1 global at 1024x1024
        """
        # Token counts for Qwen2 encoder
        TOKENS_PER_PATCH = 144  # 12x12 SAM features for 768x768
        TOKENS_PER_GLOBAL = 256  # 16x16 SAM features for 1024x1024
        TOKENS_VIEW_SEP = 1

        assert conversation.count(self.image_token) == len(
            images
        ), f"The number of image tokens in the prompt does not match the number of images: {conversation.count(self.image_token)} != {len(images)}"

        text_splits = conversation.split(self.image_token)

        all_patches_list = []
        all_global_list = []
        images_seq_mask = []
        tokenized_str = []
        images_spatial_crop = []
        num_image_tokens_list = []

        for text_sep, image in zip(text_splits, images):
            # Tokenize the text before this image
            tokenized_sep = self.encode(text_sep, bos=False, eos=False)
            tokenized_str += tokenized_sep
            images_seq_mask += [False] * len(tokenized_sep)

            # Process global view: pad to base_size x base_size (1024x1024)
            global_view = ImageOps.pad(
                image,
                (base_size, base_size),
                color=tuple(int(x * 255) for x in self.image_transform.mean),
            )
            global_tensor = self.image_transform(global_view).astype(mx.bfloat16)
            all_global_list.append(global_tensor)

            # Process local patches using dynamic resolution
            if cropping and min_patches > 0:
                # Use dynamic_preprocess to split image into patches
                patches, (rows, cols) = dynamic_preprocess(
                    image,
                    min_num=min_patches,
                    max_num=max_patches,
                    image_size=image_size,  # 768x768 patches
                    use_thumbnail=False,
                )
                num_patches = len(patches)

                # Transform each patch
                patch_tensors = []
                for patch in patches:
                    patch_tensor = self.image_transform(patch).astype(mx.bfloat16)
                    patch_tensors.append(patch_tensor)

                if patch_tensors:
                    patches_stacked = mx.stack(patch_tensors, axis=0)
                    all_patches_list.append(patches_stacked)

                images_spatial_crop.append([rows, cols])
            else:
                # No patches, only global view
                num_patches = 0
                images_spatial_crop.append([0, 0])

            # Calculate number of image tokens for this image
            # Order: [local_patches, global_view, view_separator]
            num_image_tokens = (
                (num_patches * TOKENS_PER_PATCH) + TOKENS_PER_GLOBAL + TOKENS_VIEW_SEP
            )
            num_image_tokens_list.append(num_image_tokens)

            # Add image tokens to sequence
            tokenized_image = [self.image_token_id] * num_image_tokens
            tokenized_str += tokenized_image
            images_seq_mask += [True] * len(tokenized_image)

        # Tokenize the text after the last image
        tokenized_sep = self.encode(text_splits[-1], bos=False, eos=False)
        tokenized_str += tokenized_sep
        images_seq_mask += [False] * len(tokenized_sep)

        # Add the bos token
        bos_id = 0
        tokenized_str = [bos_id] + tokenized_str
        images_seq_mask = [False] + images_seq_mask

        images_seq_mask = mx.array(images_seq_mask)

        # Stack global images
        if len(all_global_list) == 0:
            images_ori = mx.zeros((1, 3, base_size, base_size))
            images_spatial_crop = mx.zeros((1, 2))
        else:
            images_ori = mx.stack(all_global_list, axis=0)
            images_spatial_crop = mx.array(images_spatial_crop)

        # Stack patches (or zeros if no patches)
        if all_patches_list:
            # Concatenate all patches from all images
            images_patches = mx.concatenate(all_patches_list, axis=0)
        else:
            images_patches = mx.zeros((1, 3, image_size, image_size))

        assert len(tokenized_str) == len(
            images_seq_mask
        ), f"tokenize_with_images func: tokenized_str's length {len(tokenized_str)} is not equal to images_seq_mask's length {len(images_seq_mask)}"

        return (
            tokenized_str,
            [images_patches, images_ori],
            images_seq_mask,
            images_spatial_crop,
            num_image_tokens_list[0] if num_image_tokens_list else 257,
        )

    def __call__(
        self,
        *,
        text: str = None,
        images: List[Image.Image] = None,
        inference_mode: bool = True,
        image_size: int = 768,
        base_size: int = 1024,
        cropping: bool = True,
        min_patches: int = 1,
        max_patches: int = 6,
        padding: bool = True,
        return_tensors: Literal["np", "mx", "pt"] = "mx",
        **kwargs,
    ):
        """Process text and images for DeepSeek-OCR-2.

        Args:
            text (str or List[str]): the formatted prompt(s)
            images (List[ImageType]): the list of images (one per prompt for batched inputs)
            inference_mode (bool): if True, remove the last eos token
            image_size (int): size of local patches (default 768)
            base_size (int): size of global view (default 1024)
            cropping (bool): whether to use dynamic resolution with local patches
            min_patches (int): minimum number of patches (default 1)
            max_patches (int): maximum number of patches (default 6)

        Returns:
            outputs (dict): the output of the processor,
                - input_ids (mx.array): [batch_size, N + image tokens]
                - images (List[mx.array]): [patches, global_images]
                - images_seq_mask (mx.array): mask for image token positions
                - images_spatial_crop (mx.array): patch grid dimensions
        """

        # Handle batched inputs (list of prompts with list of images)
        if isinstance(text, list):
            if images is None:
                images = [None] * len(text)

            batch_results = []
            for i, prompt in enumerate(text):
                # Each prompt has one image
                img = [images[i]] if images[i] is not None else None
                result = self.process_one(
                    prompt=prompt,
                    images=img,
                    inference_mode=inference_mode,
                    image_size=image_size,
                    base_size=base_size,
                    cropping=cropping,
                    min_patches=min_patches,
                    max_patches=max_patches,
                )
                batch_results.append(result)

            # Collate batch results
            return self._collate_batch(batch_results, padding=padding)

        # Single input case
        prepare = self.process_one(
            prompt=text,
            images=images,
            inference_mode=inference_mode,
            image_size=image_size,
            base_size=base_size,
            cropping=cropping,
            min_patches=min_patches,
            max_patches=max_patches,
        )

        return prepare

    def _collate_batch(self, batch_results: List[Dict], padding: bool = True) -> Dict:
        """Collate multiple processed results into a batch."""
        if not batch_results:
            return {}

        batch_size = len(batch_results)

        # Get max sequence length for padding
        max_seq_len = max(r["input_ids"].shape[1] for r in batch_results)

        # Pad and stack input_ids
        padded_input_ids = []
        padded_images_seq_mask = []
        for r in batch_results:
            seq_len = r["input_ids"].shape[1]
            pad_len = max_seq_len - seq_len

            if pad_len > 0:
                # Pad input_ids on the left
                input_ids = mx.concatenate(
                    [
                        mx.full((1, pad_len), self.pad_id, dtype=r["input_ids"].dtype),
                        r["input_ids"],
                    ],
                    axis=1,
                )
                # Pad images_seq_mask on the left with False
                seq_mask = mx.concatenate(
                    [mx.zeros((1, pad_len), dtype=mx.bool_), r["images_seq_mask"]],
                    axis=1,
                )
            else:
                input_ids = r["input_ids"]
                seq_mask = r["images_seq_mask"]

            padded_input_ids.append(input_ids)
            padded_images_seq_mask.append(seq_mask)

        # Stack into batch
        input_ids = mx.concatenate(padded_input_ids, axis=0)
        images_seq_mask = mx.concatenate(padded_images_seq_mask, axis=0)

        # Combine images: [patches, global_images]
        all_patches = []
        all_global_images = []
        all_spatial_crops = []

        for r in batch_results:
            patches, global_img = r["images"]
            # Only add non-zero patches
            if mx.sum(patches).item() != 0:
                all_patches.append(patches)
            all_global_images.append(global_img)
            all_spatial_crops.append(r["images_spatial_crop"])

        # Stack patches and global images
        if all_patches:
            combined_patches = mx.concatenate(all_patches, axis=0)
        else:
            combined_patches = mx.zeros((1, 3, 1024, 1024))

        combined_global_images = mx.concatenate(all_global_images, axis=0)
        combined_spatial_crops = mx.concatenate(all_spatial_crops, axis=0)

        return {
            "input_ids": input_ids,
            "attention_mask": input_ids != self.pad_id,
            "images": [combined_patches, combined_global_images],
            "images_seq_mask": images_seq_mask,
            "images_spatial_crop": combined_spatial_crops,
        }


# Install a composable AutoProcessor patch for DeepSeek-OCR-2
from ..base import install_auto_processor_patch

install_auto_processor_patch("deepseekocr_2", DeepseekOCR2Processor)



# SmolVLM uses the same configuration structure as Idefics3
# We define our own configs with defaults for SmolVLM
from dataclasses import dataclass, field
from typing import List, Optional

from ..base import BaseModelConfig


@dataclass
class TextConfig(BaseModelConfig):
    model_type: str = "smolvlm"
    hidden_size: int = 4096
    intermediate_size: int = 11008
    num_attention_heads: Optional[int] = None
    rms_norm_eps: float = 1e-5
    vocab_size: int = 49152
    num_key_value_heads: Optional[int] = None
    head_dim: Optional[int] = None
    rope_theta: float = 1000000.0
    num_hidden_layers: int = 32
    rope_traditional: bool = False
    max_position_embeddings: int = 4096
    tie_word_embeddings: bool = False

    def __post_init__(self):
        if self.num_attention_heads is None:
            if self.head_dim is not None and self.head_dim > 0:
                self.num_attention_heads = self.hidden_size // self.head_dim
            else:
                self.num_attention_heads = 32
        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads


@dataclass
class VisionConfig(BaseModelConfig):
    model_type: str = "siglip_vision_model"
    hidden_size: Optional[int] = None
    num_attention_heads: Optional[int] = None
    patch_size: int = 14
    num_hidden_layers: Optional[int] = None
    intermediate_size: Optional[int] = None
    image_size: int = 384
    num_channels: int = 3
    layer_norm_eps: float = 1e-6

    def __post_init__(self):
        if self.hidden_size is None:
            self.hidden_size = 1152

        if self.num_attention_heads is None:
            if self.hidden_size % 64 == 0:
                self.num_attention_heads = self.hidden_size // 64
            else:
                self.num_attention_heads = 16

        if self.num_hidden_layers is None:
            # SmolVLM2 variants use either the 12-layer (500M) or 27-layer (2.2B) vision tower.
            self.num_hidden_layers = 12 if self.hidden_size <= 768 else 27

        if self.intermediate_size is None:
            if self.hidden_size == 768:
                self.intermediate_size = 3072
            elif self.hidden_size == 1152:
                self.intermediate_size = 4304
            else:
                self.intermediate_size = self.hidden_size * 4


@dataclass
class ModelConfig(BaseModelConfig):
    text_config: "TextConfig" = field(default_factory=lambda: TextConfig())
    vision_config: "VisionConfig" = field(default_factory=lambda: VisionConfig())
    model_type: str = "smolvlm"
    ignore_index: int = -100
    vocab_size: int = 49152
    scale_factor: int = 2
    image_token_id: int = 49153
    image_token_index: Optional[int] = None
    eos_token_id: Optional[List[int]] = None

    def __post_init__(self):
        if self.image_token_index is None:
            self.image_token_index = self.image_token_id


__all__ = ["ModelConfig", "TextConfig", "VisionConfig"]



import math
from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from mlx_vlm.models.qwen3_omni_moe.config import AudioConfig


def _get_feat_extract_output_lengths(input_lengths):
    input_lengths_leave = input_lengths % 100
    feat_lengths = (input_lengths_leave - 1) // 2 + 1
    output_lengths = (
        ((feat_lengths - 1) // 2 + 1 - 1) // 2 + 1 + (input_lengths // 100) * 13
    )
    return output_lengths


class Attention(nn.Module):
    def __init__(self, config: AudioConfig):
        super().__init__()
        self.embed_dim = config.d_model
        self.num_heads = config.encoder_attention_heads
        self.head_dim = self.embed_dim // self.num_heads
        self.scaling = self.head_dim**-0.5

        self.k_proj = nn.Linear(config.d_model, config.d_model, bias=True)
        self.v_proj = nn.Linear(config.d_model, config.d_model, bias=True)
        self.q_proj = nn.Linear(config.d_model, config.d_model, bias=True)
        self.out_proj = nn.Linear(config.d_model, config.d_model, bias=True)

    def __call__(
        self,
        hidden_states: mx.array,
        cu_seqlens: mx.array,
        attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        seq_length = hidden_states.shape[0]

        query_states = self.q_proj(hidden_states).reshape(
            seq_length, self.num_heads, -1
        )
        key_states = self.k_proj(hidden_states).reshape(seq_length, self.num_heads, -1)
        value_states = self.v_proj(hidden_states).reshape(
            seq_length, self.num_heads, -1
        )

        query_states = query_states.transpose(1, 0, 2)[None]
        key_states = key_states.transpose(1, 0, 2)[None]
        value_states = value_states.transpose(1, 0, 2)[None]

        lengths = (cu_seqlens[1:] - cu_seqlens[:-1]).tolist()
        if len(lengths) == 0:
            lengths = [seq_length]

        attn_outputs = []
        offset = 0
        for length in lengths:
            if length <= 0:
                continue
            end = offset + length
            q_chunk = query_states[:, :, offset:end]
            k_chunk = key_states[:, :, offset:end]
            v_chunk = value_states[:, :, offset:end]
            attn_weights = (q_chunk @ k_chunk.swapaxes(-2, -1)) * self.scaling
            attn_weights = mx.softmax(attn_weights, axis=-1)
            attn_output = attn_weights @ v_chunk
            attn_outputs.append(attn_output)
            offset = end

        attn_output = (
            mx.concatenate(attn_outputs, axis=2)
            if attn_outputs
            else mx.zeros_like(query_states)
        )
        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(seq_length, -1)
        attn_output = self.out_proj(attn_output)

        return attn_output


class AudioEncoderLayer(nn.Module):
    def __init__(self, config: AudioConfig, idx: int):
        super().__init__()
        self.embed_dim = config.d_model
        self.self_attn = Attention(config)
        self.self_attn_layer_norm = nn.LayerNorm(config.d_model)
        self.fc1 = nn.Linear(config.d_model, config.encoder_ffn_dim)
        self.fc2 = nn.Linear(config.encoder_ffn_dim, config.d_model)
        self.final_layer_norm = nn.LayerNorm(config.d_model)

        if config.activation_function == "gelu":
            self.activation_fn = nn.gelu
        else:
            raise ValueError(f"Unsupported activation: {config.activation_function}")

    def __call__(
        self,
        hidden_states: mx.array,
        cu_seqlens: mx.array,
        attention_mask: Optional[mx.array] = None,
    ) -> mx.array:
        residual = hidden_states
        hidden_states = self.self_attn_layer_norm(hidden_states)
        hidden_states = self.self_attn(
            hidden_states=hidden_states,
            cu_seqlens=cu_seqlens,
            attention_mask=attention_mask,
        )
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.final_layer_norm(hidden_states)
        hidden_states = self.fc1(hidden_states)
        hidden_states = self.activation_fn(hidden_states)
        hidden_states = self.fc2(hidden_states)
        hidden_states = residual + hidden_states

        return hidden_states


class SinusoidsPositionEmbedding(nn.Module):
    def __init__(self, length, channels, max_timescale=10000):
        super().__init__()
        if channels % 2 != 0:
            raise ValueError("SinusoidsPositionEmbedding needs even channels input")
        log_timescale_increment = np.log(max_timescale) / (channels // 2 - 1)
        inv_timescales = np.exp(
            -log_timescale_increment * np.arange(channels // 2, dtype=np.float32)
        )
        scaled_time = np.arange(length)[:, np.newaxis] * inv_timescales[np.newaxis, :]
        positional_embedding = np.concatenate(
            [np.sin(scaled_time), np.cos(scaled_time)], axis=1
        )
        self.positional_embedding = positional_embedding

    def __call__(self, seqlen: int) -> mx.array:
        return mx.array(self.positional_embedding[:seqlen, :])


class AudioModel(nn.Module):
    def __init__(self, config: AudioConfig):
        super().__init__()

        self.dropout = config.dropout
        embed_dim = config.d_model
        self.num_mel_bins = config.num_mel_bins
        self.max_source_positions = config.max_source_positions
        self.embed_scale = math.sqrt(embed_dim) if config.scale_embedding else 1.0
        self.n_window = config.n_window
        self.n_window_infer = config.n_window_infer
        self.conv_chunksize = config.conv_chunksize

        self.positional_embedding = SinusoidsPositionEmbedding(
            self.max_source_positions, embed_dim
        )
        self.layers = [
            AudioEncoderLayer(config, idx) for idx in range(config.encoder_layers)
        ]
        self.ln_post = nn.LayerNorm(config.d_model)

        self.conv2d1 = nn.Conv2d(1, config.downsample_hidden_size, 3, 2, padding=1)
        self.conv2d2 = nn.Conv2d(
            config.downsample_hidden_size,
            config.downsample_hidden_size,
            3,
            2,
            padding=1,
        )
        self.conv2d3 = nn.Conv2d(
            config.downsample_hidden_size,
            config.downsample_hidden_size,
            3,
            2,
            padding=1,
        )
        self.conv_out = nn.Linear(
            config.downsample_hidden_size
            * ((((config.num_mel_bins + 1) // 2 + 1) // 2 + 1) // 2),
            config.d_model,
            bias=False,
        )
        self.proj1 = nn.Linear(config.d_model, config.d_model)

        if config.activation_function == "gelu":
            self.act = nn.gelu
        else:
            raise ValueError(f"Unsupported activation: {config.activation_function}")

        self.proj2 = nn.Linear(config.d_model, config.output_dim)

    def __call__(
        self,
        input_features: mx.array,
        feature_lens: Optional[mx.array] = None,
        aftercnn_lens: Optional[mx.array] = None,
    ):
        if feature_lens is None:
            feature_lens = mx.array([input_features.shape[-1]], dtype=mx.int32)

        aftercnn_lens = _get_feat_extract_output_lengths(feature_lens)
        feature_lens_np = np.array(feature_lens).astype(np.int32)
        n_window_step = self.n_window * 2
        chunk_num = np.ceil(feature_lens_np / n_window_step).astype(np.int32)

        chunk_lengths_list = []
        tail_chunk_info = []
        cumsum = 0
        for sample_idx, num_chunks in enumerate(chunk_num.tolist()):
            num_int = int(num_chunks)
            chunk_lengths_list.extend([n_window_step] * num_int)
            if num_int > 0:
                tail_chunk_info.append((cumsum + num_int - 1, sample_idx))
            cumsum += num_int

        for tail_idx, sample_idx in tail_chunk_info:
            remainder = feature_lens_np[sample_idx] % n_window_step
            if remainder == 0:
                remainder = n_window_step
            chunk_lengths_list[tail_idx] = int(remainder)

        chunk_lengths = mx.array(chunk_lengths_list, dtype=mx.int32)

        total_chunks = len(chunk_lengths_list)
        max_chunk_len = int(chunk_lengths.max())
        padded_feature = mx.zeros(
            (total_chunks, self.num_mel_bins, max_chunk_len), dtype=input_features.dtype
        )

        start_idx = 0
        for i, chunk_len in enumerate(chunk_lengths_list):
            end_idx = start_idx + chunk_len
            padded_feature[i, :, :chunk_len] = input_features[:, start_idx:end_idx]
            start_idx = end_idx

        padded_feature = padded_feature[:, None, :, :]

        feature_lens_after_cnn = _get_feat_extract_output_lengths(feature_lens)
        max_len_after_cnn = int(feature_lens_after_cnn.max())
        padded_mask_after_cnn = mx.zeros(
            (total_chunks, max_len_after_cnn), dtype=mx.bool_
        )
        for i, length in enumerate(feature_lens_after_cnn):
            padded_mask_after_cnn[i, : int(length)] = True

        padded_embeds = []
        for i in range(0, total_chunks, self.conv_chunksize):
            end_idx = min(i + self.conv_chunksize, total_chunks)
            chunk = padded_feature[i:end_idx]
            chunk = chunk.transpose(0, 2, 3, 1)
            padded_embed = nn.gelu(self.conv2d1(chunk))
            padded_embed = nn.gelu(self.conv2d2(padded_embed))
            padded_embed = nn.gelu(self.conv2d3(padded_embed))
            padded_embeds.append(padded_embed)

        padded_embed = mx.concatenate(padded_embeds, axis=0)
        b, h, w, c = padded_embed.shape
        padded_embed = padded_embed.transpose(0, 2, 3, 1).reshape(b, w, c * h)
        padded_embed = self.conv_out(padded_embed)

        seq_len = padded_embed.shape[1]
        positional_embedding = self.positional_embedding(seq_len)[None]
        padded_embed = padded_embed + positional_embedding

        linear_indices = []
        for i in range(total_chunks):
            mask_array = np.array(padded_mask_after_cnn[i])
            chunk_indices = np.where(mask_array)[0]
            linear_indices.extend([i * seq_len + idx for idx in chunk_indices])

        padded_embed_flat = padded_embed.reshape(-1, padded_embed.shape[-1])
        hidden_states = mx.take(
            padded_embed_flat,
            mx.array(np.array(linear_indices, dtype=np.int32)),
            axis=0,
        )

        cu_chunk_lens = []
        window_aftercnn = max_len_after_cnn * (
            self.n_window_infer // (self.n_window * 2)
        )
        for cnn_len in feature_lens_after_cnn:
            cnn_len_int = int(cnn_len)
            num_windows = cnn_len_int // window_aftercnn
            cu_chunk_lens.extend([window_aftercnn] * num_windows)
            remainder = cnn_len_int % window_aftercnn
            if remainder != 0:
                cu_chunk_lens.append(remainder)

        cu_seqlens = mx.cumsum(mx.array(cu_chunk_lens, dtype=mx.int32), axis=0)
        cu_seqlens = mx.pad(cu_seqlens, (1, 0), constant_values=0)

        for i, encoder_layer in enumerate(self.layers):
            hidden_states = encoder_layer(
                hidden_states,
                cu_seqlens,
            )
            if i % 2 == 0:
                mx.eval(hidden_states)

        hidden_states = self.ln_post(hidden_states)
        hidden_states = self.proj1(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.proj2(hidden_states)

        return hidden_states

    def sanitize(self, weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if "audio_tower.conv2d" in k and "weight" in k:
                sanitized_weights[k] = v.transpose(0, 2, 3, 1)
            else:
                sanitized_weights[k] = v

        return sanitized_weights



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from transformers import AutoImageProcessor, AutoProcessor

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .processing_kimi_vl import KimiVLImageProcessor, KimiVLProcessor
from .vision import VisionModel

# Register custom processor classes for kimi_vl model type
try:
    MODEL_TYPE = "kimi_vl"
    AutoImageProcessor.register(
        MODEL_TYPE, slow_image_processor_class=KimiVLImageProcessor
    )
    AutoProcessor.register(MODEL_TYPE, KimiVLProcessor)
except Exception:
    raise Exception("Failed to register kimi_vl processor")


class KimiVLMultiModalProjector(nn.Module):

    def __init__(self, config: ModelConfig):
        super().__init__()

        self.hidden_size = (
            config.vision_config.hidden_size
            * config.vision_config.merge_kernel_size[0]
            * config.vision_config.merge_kernel_size[1]
        )

        self.pre_norm = nn.LayerNorm(config.vision_config.hidden_size, eps=1e-05)
        self.linear_1 = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.act = nn.GELU()
        self.linear_2 = nn.Linear(
            self.hidden_size, config.text_config.hidden_size, bias=True
        )

    def __call__(self, image_features: list[mx.array]) -> mx.array:
        image_features = mx.concatenate(image_features, axis=0)
        h = self.pre_norm(image_features).reshape(-1, self.hidden_size)
        h = self.linear_1(h)
        h = self.act(h)
        h = self.linear_2(h)
        return h


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.model_type = config.model_type
        self.config = config

        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.multi_modal_projector = KimiVLMultiModalProjector(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        image_grid_thw = kwargs.pop("image_grid_hws", None)
        video_grid_thw = kwargs.pop("video_grid_hws", None)
        image_token_id = kwargs.pop("image_token_id", None)
        grid_thw = image_grid_thw if image_grid_thw is not None else video_grid_thw

        if pixel_values is None:
            inputs_embeds = self.language_model.embed_tokens(input_ids)
            return InputEmbeddingsFeatures(inputs_embeds=inputs_embeds)

        inputs_embeds = self.language_model.embed_tokens(input_ids)

        hidden_state = self.vision_tower(
            pixel_values.transpose(0, 2, 3, 1),
            output_hidden_states=True,
            grid_thw=grid_thw,
        )

        image_features = self.multi_modal_projector(hidden_state)

        final_inputs_embeds = self._prepare_inputs_for_multimodal(
            image_features,
            inputs_embeds,
            input_ids,
            image_token_id=image_token_id,
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    def _prepare_inputs_for_multimodal(
        self,
        image_features,
        inputs_embeds,
        input_ids,
        image_token_id=None,
    ):
        candidate_token_ids = []
        for token_id in [
            image_token_id,
            self.config.image_token_index,
            getattr(self.config, "media_placeholder_token_id", None),
        ]:
            if token_id is None:
                continue
            if isinstance(token_id, mx.array):
                if token_id.size == 0:
                    continue
                token_id = token_id.item()
            token_id = int(token_id)
            if token_id not in candidate_token_ids:
                candidate_token_ids.append(token_id)

        image_mask = mx.zeros(input_ids.shape, dtype=mx.bool_)
        for token_id in candidate_token_ids:
            image_mask = mx.logical_or(image_mask, input_ids == token_id)

        # Positions of <image> tokens in input_ids, assuming batch size is 1
        image_positions = np.where(np.array(image_mask))[1].tolist()
        num_image_tokens = len(image_positions)
        num_image_features = image_features.shape[0]
        if num_image_tokens != num_image_features:
            raise ValueError(
                "Number of image placeholder tokens does not match extracted image features: "
                f"{num_image_tokens} tokens for {num_image_features} features. "
                f"Candidate token IDs: {candidate_token_ids}."
            )

        inputs_embeds[:, image_positions, :] = image_features

        return inputs_embeds

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        cache=None,
        **kwargs,
    ):

        input_embeddings_features = self.get_input_embeddings(
            input_ids, pixel_values, **kwargs
        )
        logits = self.language_model(
            inputs=input_ids,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
        )
        return logits

    def sanitize(self, weights):
        return {
            k.replace("encoder.", "") if "vision_tower" in k else k: v
            for k, v in weights.items()
        }



from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from ..base import (
    LanguageModelOutput,
    create_attention_mask,
    scaled_dot_product_attention,
)
from ..cache import KVCache
from .config import ModelConfig, TextConfig


class Molmo2Embedding(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        num_new_embeddings: int,
        features: int,
    ):
        super().__init__()
        self.embedding = mx.zeros((num_embeddings, features))
        self.new_embedding = mx.zeros((num_new_embeddings, features))

    def __call__(self, x: mx.array) -> mx.array:
        return mx.concatenate([self.embedding, self.new_embedding], axis=0)[x]


class LanguageModelMLP(nn.Module):
    def __init__(self, input_dim: int, intermediate_size: int):
        super().__init__()
        self.ff_proj = nn.Linear(input_dim, intermediate_size * 2, bias=False)
        self.ff_out = nn.Linear(intermediate_size, input_dim, bias=False)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.ff_proj(x)
        x, gate = mx.split(x, 2, axis=-1)
        x = nn.silu(gate) * x
        return self.ff_out(x)


class Molmo2Attention(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.num_heads = config.num_attention_heads
        self.num_key_value_heads = config.num_key_value_heads
        self.head_dim = config.head_dim
        self.scale = self.head_dim**-0.5

        self.fused_dims = (
            config.num_attention_heads * config.head_dim,
            config.head_dim * config.num_key_value_heads,
            config.head_dim * config.num_key_value_heads,
        )

        self.att_proj = nn.Linear(
            config.hidden_size,
            sum(self.fused_dims),
            bias=config.qkv_bias,
        )

        self.q_norm = nn.RMSNorm(dims=config.head_dim, eps=config.layer_norm_eps)
        self.k_norm = nn.RMSNorm(dims=config.head_dim, eps=config.layer_norm_eps)

        self.attn_out = nn.Linear(
            config.head_dim * config.num_attention_heads,
            config.hidden_size,
            bias=False,
        )

        self.rotary_emb = nn.RoPE(self.head_dim, base=config.rope_theta)

    def __call__(
        self,
        hidden_states: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        batch_size, seq_len, _ = hidden_states.shape

        qkv = self.att_proj(hidden_states)
        q, k, v = mx.split(
            qkv,
            [self.fused_dims[0], self.fused_dims[0] + self.fused_dims[1]],
            axis=-1,
        )

        q = self.q_norm(q.reshape(batch_size, seq_len, self.num_heads, self.head_dim))
        k = self.k_norm(
            k.reshape(batch_size, seq_len, self.num_key_value_heads, self.head_dim)
        )
        v = v.reshape(batch_size, seq_len, self.num_key_value_heads, self.head_dim)

        q = q.transpose(0, 2, 1, 3)
        k = k.transpose(0, 2, 1, 3)
        v = v.transpose(0, 2, 1, 3)

        if cache is not None:
            q = self.rotary_emb(q, offset=cache.offset)
            k = self.rotary_emb(k, offset=cache.offset)
            k, v = cache.update_and_fetch(k, v)
        else:
            q = self.rotary_emb(q)
            k = self.rotary_emb(k)

        att = scaled_dot_product_attention(q, k, v, cache, scale=self.scale, mask=mask)
        att = att.transpose(0, 2, 1, 3).reshape(batch_size, seq_len, -1)
        return self.attn_out(att)


class Molmo2DecoderLayer(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.self_attn = Molmo2Attention(config)
        self.attn_norm = nn.RMSNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.ff_norm = nn.RMSNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.mlp = LanguageModelMLP(config.hidden_size, config.intermediate_size)

    def __call__(
        self,
        hidden_states: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
    ) -> mx.array:
        residual = hidden_states
        hidden_states = self.attn_norm(hidden_states)
        hidden_states = residual + self.self_attn(hidden_states, mask, cache)

        residual = hidden_states
        hidden_states = self.ff_norm(hidden_states)
        hidden_states = residual + self.mlp(hidden_states)
        return hidden_states


class Molmo2Transformer(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config

        self.wte = Molmo2Embedding(
            config.vocab_size, config.additional_vocab_size, config.hidden_size
        )
        self.blocks = [
            Molmo2DecoderLayer(config) for _ in range(config.num_hidden_layers)
        ]
        self.ln_f = nn.RMSNorm(config.hidden_size, eps=config.layer_norm_eps)

        self.emb_drop = nn.Dropout(config.embedding_dropout)

    def __call__(
        self,
        input_ids: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache: Optional[list[KVCache]] = None,
    ) -> mx.array:
        if inputs_embeds is None:
            hidden_states = self.wte(input_ids)
        else:
            hidden_states = inputs_embeds

        if cache is None:
            cache = [None] * len(self.blocks)

        if mask is None:
            mask = create_attention_mask(hidden_states, cache)

        hidden_states = self.emb_drop(hidden_states)

        for block, c in zip(self.blocks, cache):
            hidden_states = block(hidden_states, mask, c)

        return self.ln_f(hidden_states)


class LanguageModel(nn.Module):
    def __init__(self, args: TextConfig, config: ModelConfig = None):
        super().__init__()
        self.args = args
        self.config = config
        self.model_type = args.model_type
        self.model = Molmo2Transformer(args)
        self.lm_head = nn.Linear(args.hidden_size, args.vocab_size, bias=False)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache: Optional[list[KVCache]] = None,
        **kwargs,
    ) -> LanguageModelOutput:
        hidden_states = self.model(inputs, inputs_embeds, mask, cache)
        logits = self.lm_head(hidden_states)
        return LanguageModelOutput(logits=logits)

    @staticmethod
    def sanitize(weights):
        # Remove unused precomputed rotary freqs if present.
        return {k: v for k, v in weights.items() if "rotary_emb.inv_freq" not in k}

    @property
    def layers(self):
        return self.model.blocks



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..interpolate import resize_bilinear
from .config import VisionConfig


def check_array_shape(arr):
    shape = arr.shape

    # Check if the shape has 4 dimensions
    if len(shape) != 4:
        return False

    out_channels, kH, KW, _ = shape

    # Check if out_channels is the largest, and kH and KW are the same
    if (out_channels >= kH) and (out_channels >= KW) and (kH == KW):
        return True
    else:
        return False


class Attention(nn.Module):
    def __init__(
        self,
        dims: int,
        num_heads: int,
        query_input_dims: Optional[int] = None,
        key_input_dims: Optional[int] = None,
        value_input_dims: Optional[int] = None,
        value_dims: Optional[int] = None,
        value_output_dims: Optional[int] = None,
        bias: bool = True,
    ):
        super().__init__()

        if (dims % num_heads) != 0:
            raise ValueError(
                "The input feature dimensions should be divisible by the "
                f"number of heads ({dims} % {num_heads}) != 0"
            )

        query_input_dims = query_input_dims or dims
        key_input_dims = key_input_dims or dims
        value_input_dims = value_input_dims or key_input_dims
        value_dims = value_dims or dims
        value_output_dims = value_output_dims or dims

        self.num_heads = num_heads
        head_dim = dims // num_heads
        self.scale = head_dim**-0.5

        self.q_proj = nn.Linear(query_input_dims, dims, bias=bias)
        self.k_proj = nn.Linear(key_input_dims, dims, bias=bias)
        self.v_proj = nn.Linear(value_input_dims, value_dims, bias=bias)
        self.out_proj = nn.Linear(value_dims, value_output_dims, bias=bias)

    def __call__(self, x, mask=None):
        queries = self.q_proj(x)
        keys = self.k_proj(x)
        values = self.v_proj(x)

        num_heads = self.num_heads
        B, L, D = queries.shape
        _, S, _ = keys.shape
        queries = queries.reshape(B, L, num_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, S, num_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, S, num_heads, -1).transpose(0, 2, 1, 3)

        output = mx.fast.scaled_dot_product_attention(
            queries, keys, values, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.out_proj(output)


class MLP(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.activation_fn = nn.GELU(approx="precise")
        self.fc1 = nn.Linear(config.hidden_size, config.intermediate_size, bias=True)
        self.fc2 = nn.Linear(config.intermediate_size, config.hidden_size, bias=True)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.fc1(x)
        x = self.activation_fn(x)
        x = self.fc2(x)
        return x


class EncoderLayer(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.embed_dim = config.hidden_size
        self.self_attn = Attention(
            config.hidden_size, config.num_attention_heads, bias=True
        )
        self.layer_norm1 = nn.LayerNorm(self.embed_dim, eps=config.layer_norm_eps)
        self.mlp = MLP(config)
        self.layer_norm2 = nn.LayerNorm(self.embed_dim, eps=config.layer_norm_eps)

    def __call__(self, x: mx.array, mask: Optional[mx.array] = None) -> mx.array:
        r = self.self_attn(self.layer_norm1(x), mask)
        h = x + r
        r = self.mlp(self.layer_norm2(h))
        return h + r


class Encoder(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.layers = [EncoderLayer(config) for _ in range(config.num_hidden_layers)]

    def __call__(
        self,
        x: mx.array,
        output_hidden_states: Optional[bool] = None,
        mask: Optional[mx.array] = None,
    ) -> mx.array:
        encoder_states = (x,) if output_hidden_states else None
        h = x
        for l in self.layers:
            x = l(x, mask=mask)
            if output_hidden_states:
                encoder_states = encoder_states + (x,)

            h = x

        return (h, encoder_states)


def gaussian_blur_axis(image, sigma, axis):
    """
    Applies a 1D Gaussian blur along the given axis.
    This version works for arrays with any number of dimensions.
    """
    radius = int(3 * sigma)
    if radius < 1:
        return image
    x = mx.arange(-radius, radius + 1)
    kernel = mx.exp(-(x**2) / (2 * sigma**2))
    kernel = kernel / mx.sum(kernel)

    # MLX doesn't have a direct apply_along_axis equivalent,
    # so we'll implement the convolution differently based on the axis

    # Helper function to apply 1D convolution along specific axis
    def conv_1d(array, kernel, axis):
        # Reshape kernel to broadcast along the right dimensions
        kernel_shape = [1] * image.ndim
        kernel_shape[axis] = len(kernel)
        kernel_reshaped = kernel.reshape(kernel_shape)

        # Pad the array
        pad_width = [(0, 0)] * image.ndim
        pad_width[axis] = (radius, radius)
        padded = mx.pad(array, pad_width, mode="edge")

        # Perform convolution via sliding window sum
        result = mx.zeros_like(array)
        slices = [slice(None)] * padded.ndim

        for i in range(2 * radius + 1):
            slices[axis] = slice(i, i + array.shape[axis])
            result = result + padded[tuple(slices)] * kernel_reshaped

        return result

    return conv_1d(image, kernel, axis)


class VisionEmbeddings(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.config = config
        self.embed_dim = config.hidden_size
        self.image_size = config.image_size
        self.patch_size = config.patch_size

        self.patch_embedding = nn.Conv2d(
            config.num_channels,
            config.hidden_size,
            kernel_size=self.patch_size,
            stride=self.patch_size,
        )

        self.num_patches = (self.image_size // self.patch_size) ** 2
        self.num_positions = self.num_patches
        self.position_embedding = nn.Embedding(self.num_positions, self.embed_dim)

    @staticmethod
    def resize_positional_embeddings(
        positional_embeddings: mx.array,
        spatial_shapes: mx.array,
        max_length: int,
    ) -> mx.array:
        """
        Resize positional embeddings to image-specific size and pad to a fixed size.

        Args:
            positional_embeddings (`torch.Tensor`):
                Position embeddings of shape (height, width, embed_dim)
            spatial_shapes (`torch.LongTensor`):
                Spatial shapes of shape (batch_size, 2) to resize the positional embeddings to
            max_length (`int`):
                Maximum length of the positional embeddings to pad resized positional embeddings to

        Returns:
            `torch.Tensor`: Embeddings of shape (batch_size, max_length, embed_dim)
        """
        batch_size = spatial_shapes.shape[0]
        embed_dim = positional_embeddings.shape[-1]
        source_dtype = positional_embeddings.dtype

        resulted_positional_embeddings = mx.zeros(
            (batch_size, max_length, embed_dim)
        ).astype(source_dtype)

        # (height, width, embed_dim) -> (1, embed_dim, height, width) for interpolation
        positional_embeddings = positional_embeddings.transpose(2, 0, 1).reshape(
            1, embed_dim, -1
        )

        # Upcast to float32 on CPU because antialias is not supported for bfloat16/float16 on CPU
        if positional_embeddings.device.type == "cpu":
            positional_embeddings = positional_embeddings.astype(mx.float32)

        for i in range(batch_size):
            # (1, dim, height, width) -> (1, dim, target_height, target_width)
            height, width = spatial_shapes[i]
            # Then upsample width dimension
            resized_embeddings = resize_bilinear(
                positional_embeddings,
                (height, width),
                align_corners=False,
                antialias=True,
            )

            # (1, dim, target_height, target_width) -> (target_height * target_width, dim)
            resized_embeddings = resized_embeddings.reshape(
                embed_dim, height * width
            ).transpose(0, 1)

            # Cast to original dtype
            resized_embeddings = resized_embeddings.astype(source_dtype)

            resulted_positional_embeddings[i, : height * width] = resized_embeddings
            resulted_positional_embeddings[i, height * width :] = resized_embeddings[0]

        return resulted_positional_embeddings

    def __call__(
        self, x: mx.array, spatial_shapes: Optional[mx.array] = None
    ) -> mx.array:
        batch_size = x.shape[0]
        patch_embeddings = self.patch_embedding(x)
        patch_embeddings = mx.flatten(patch_embeddings, start_axis=1, end_axis=2)
        if spatial_shapes is None:
            position_ids = mx.array(np.arange(self.num_positions)[None, :])
            embeddings = patch_embeddings
            embeddings += self.position_embedding(position_ids)

        else:
            # Get positional resized and padded positional embeddings
            positional_embeddings = self.position_embedding.weight.reshape(
                self.position_embedding_size, self.position_embedding_size, -1
            )

            resized_positional_embeddings = self.resize_positional_embeddings(
                positional_embeddings, spatial_shapes, max_length=x.shape[1]
            )

            # Add positional embeddings to patch embeddings
            embeddings = patch_embeds + resized_positional_embeddings
        return embeddings


class SigLipVisionModel(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()

        self.embeddings = VisionEmbeddings(config)
        self.encoder = Encoder(config)
        self.post_layernorm = nn.LayerNorm(config.hidden_size)

    def __call__(
        self,
        x: mx.array,
        spatial_shapes: mx.array,
        output_hidden_states: Optional[bool] = None,
    ) -> mx.array:
        x = self.embeddings(x, spatial_shapes)
        x = x.astype(self.embeddings.patch_embedding.weight.dtype)
        encoder_outputs = self.encoder(
            x=x, output_hidden_states=output_hidden_states, mask=None
        )
        pooler_output = self.post_layernorm(encoder_outputs[0])
        return pooler_output, x, encoder_outputs[-1]


class VisionModel(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.model_type = config.model_type
        if self.model_type not in ["siglip_vision_model"]:
            raise ValueError(f"Unsupported model type: {self.model_type}")

        self.vision_model = SigLipVisionModel(config)

    def __call__(
        self,
        x: mx.array,
        spatial_shapes: Optional[mx.array] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> mx.array:
        return self.vision_model(x, spatial_shapes, output_hidden_states)

    def sanitize(self, weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if "position_ids" in k:
                # Remove unused position_ids
                continue
            elif "patch_embedding.weight" in k:
                # PyTorch conv2d weight tensors have shape:
                #   [out_channels, in_channels, kH, KW]
                # MLX conv2d expects the weight be of shape:
                #   [out_channels, kH, KW, in_channels]
                if check_array_shape(v):
                    sanitized_weights[k] = v
                else:
                    sanitized_weights[k] = v.transpose(0, 2, 3, 1)
            else:
                sanitized_weights[k] = v

        return sanitized_weights



from typing import Optional

import mlx.core as mx
import mlx.nn as nn

from .config import VisionConfig


def check_array_shape(arr):
    shape = arr.shape

    if len(shape) == 4:
        out_channels, kH, KW, _ = shape
        return (out_channels >= kH) and (out_channels >= KW) and (kH == KW)
    elif len(shape) == 5:
        out_channels, kT, kH, KW, _ = shape
        return (out_channels >= kH) and (out_channels >= KW) and (kH == KW)
    else:
        return False


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return mx.concatenate([-x2, x1], axis=-1)


def apply_rotary_pos_emb_vision(
    q: mx.array, k: mx.array, cos: mx.array, sin: mx.array
) -> tuple:
    orig_q_dtype = q.dtype
    orig_k_dtype = k.dtype
    q, k = q.astype(mx.float32), k.astype(mx.float32)
    cos = mx.expand_dims(cos, axis=-2).astype(mx.float32)
    sin = mx.expand_dims(sin, axis=-2).astype(mx.float32)
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    q_embed = q_embed.astype(orig_q_dtype)
    k_embed = k_embed.astype(orig_k_dtype)
    return q_embed, k_embed


class GlmOcrVisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.dim = dim
        self.theta = theta

    def __call__(self, seqlen: int) -> mx.array:
        inv_freq = 1.0 / (
            self.theta ** (mx.arange(0, self.dim, 2, dtype=mx.float32) / self.dim)
        )
        seq = mx.arange(seqlen, dtype=inv_freq.dtype)
        freqs = mx.outer(seq, inv_freq)
        return freqs


class GlmOcrVisionPatchEmbed(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.config = config
        self.patch_size = config.patch_size
        self.temporal_patch_size = config.temporal_patch_size
        self.in_channels = config.in_channels
        self.embed_dim = config.hidden_size

        kernel_size = [self.temporal_patch_size, self.patch_size, self.patch_size]
        self.proj = nn.Conv3d(
            self.in_channels,
            self.embed_dim,
            kernel_size=kernel_size,
            stride=kernel_size,
            bias=True,
        )

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = hidden_states.reshape(
            -1,
            self.in_channels,
            self.temporal_patch_size,
            self.patch_size,
            self.patch_size,
        ).moveaxis(1, 4)

        hidden_states = self.proj(hidden_states)
        hidden_states = hidden_states.reshape(-1, self.embed_dim)
        return hidden_states


class GlmOcrVisionPatchMerger(nn.Module):
    def __init__(
        self, dim: int, context_dim: int, hidden_act: str, bias: bool = False
    ) -> None:
        super().__init__()
        self.proj = nn.Linear(dim, dim, bias=bias)
        self.post_projection_norm = nn.LayerNorm(dim)
        self.gate_proj = nn.Linear(dim, context_dim, bias=bias)
        self.up_proj = nn.Linear(dim, context_dim, bias=bias)
        self.down_proj = nn.Linear(context_dim, dim, bias=bias)

    def __call__(self, x: mx.array) -> mx.array:
        x = self.proj(x)
        x = nn.gelu(self.post_projection_norm(x))
        return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class GlmOcrVisionAttention(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.config = config
        self.dim = config.hidden_size
        self.num_heads = config.num_heads
        self.head_dim = self.dim // self.num_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Linear(
            config.hidden_size, config.hidden_size * 3, bias=config.attention_bias
        )
        self.proj = nn.Linear(
            config.hidden_size, config.hidden_size, bias=config.attention_bias
        )

        self.q_norm = nn.RMSNorm(self.head_dim, eps=config.rms_norm_eps)
        self.k_norm = nn.RMSNorm(self.head_dim, eps=config.rms_norm_eps)

    def __call__(
        self,
        hidden_states: mx.array,
        cu_seqlens: mx.array,
        position_embeddings: tuple,
    ) -> mx.array:
        seq_length = hidden_states.shape[0]

        qkv = self.qkv(hidden_states)
        qkv = qkv.reshape(seq_length, 3, self.num_heads, -1)
        qkv = qkv.transpose(1, 0, 2, 3)
        q, k, v = mx.split(qkv, 3, axis=0)
        q = q.squeeze(0)
        k = k.squeeze(0)
        v = v.squeeze(0)

        q = self.q_norm(q)
        k = self.k_norm(k)

        cos, sin = position_embeddings
        q, k = apply_rotary_pos_emb_vision(q, k, cos, sin)

        q = q.transpose(1, 0, 2)[None, ...]
        k = k.transpose(1, 0, 2)[None, ...]
        v = v.transpose(1, 0, 2)[None, ...]

        lengths = (cu_seqlens[1:] - cu_seqlens[:-1]).tolist()
        split_indices = []
        cumsum = 0
        for i, length in enumerate(lengths[:-1]):
            cumsum += length
            split_indices.append(cumsum)

        q_splits = mx.split(q, split_indices, axis=2)
        k_splits = mx.split(k, split_indices, axis=2)
        v_splits = mx.split(v, split_indices, axis=2)

        attn_outputs = []
        for q_chunk, k_chunk, v_chunk in zip(q_splits, k_splits, v_splits):
            output = mx.fast.scaled_dot_product_attention(
                q_chunk, k_chunk, v_chunk, scale=self.scale, mask=None
            )
            attn_outputs.append(output)

        attn_output = mx.concatenate(attn_outputs, axis=2)
        # Transpose from (batch, heads, seq, head_dim) to (batch, seq, heads, head_dim)
        # then reshape to (seq, hidden_size)
        attn_output = attn_output.transpose(0, 2, 1, 3).reshape(seq_length, -1)
        attn_output = self.proj(attn_output)
        return attn_output


class GlmOcrVisionMLP(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.gate_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=config.attention_bias
        )
        self.up_proj = nn.Linear(
            config.hidden_size, config.intermediate_size, bias=config.attention_bias
        )
        self.down_proj = nn.Linear(
            config.intermediate_size, config.hidden_size, bias=config.attention_bias
        )

    def __call__(self, x: mx.array) -> mx.array:
        return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class GlmOcrVisionBlock(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.norm1 = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.norm2 = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.attn = GlmOcrVisionAttention(config)
        self.mlp = GlmOcrVisionMLP(config)

    def __call__(
        self, hidden_states: mx.array, cu_seqlens: mx.array, position_embeddings: tuple
    ) -> mx.array:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            position_embeddings=position_embeddings,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class VisionModel(nn.Module):
    def __init__(self, config: VisionConfig) -> None:
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.spatial_merge_size = config.spatial_merge_size
        self.patch_size = config.patch_size

        self.patch_embed = GlmOcrVisionPatchEmbed(config)

        head_dim = config.hidden_size // config.num_heads
        self.rotary_pos_emb = GlmOcrVisionRotaryEmbedding(head_dim // 2)

        self.blocks = [GlmOcrVisionBlock(config) for _ in range(config.depth)]

        self.merger = GlmOcrVisionPatchMerger(
            dim=config.out_hidden_size,
            context_dim=config.out_hidden_size * config.in_channels,
            hidden_act=config.hidden_act,
        )

        self.downsample = nn.Conv2d(
            in_channels=config.hidden_size,
            out_channels=config.out_hidden_size,
            kernel_size=config.spatial_merge_size,
            stride=config.spatial_merge_size,
            bias=True,
        )

        self.post_layernorm = nn.RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def rot_pos_emb(self, grid_thw: mx.array):
        pos_ids = []

        for t, h, w in grid_thw.tolist():
            hpos_ids = mx.expand_dims(mx.arange(h), 1)
            hpos_ids = mx.repeat(hpos_ids, w, axis=1)
            hpos_ids = hpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            hpos_ids = mx.transpose(hpos_ids, (0, 2, 1, 3))
            hpos_ids = hpos_ids.flatten()

            wpos_ids = mx.expand_dims(mx.arange(w), 0)
            wpos_ids = mx.repeat(wpos_ids, h, axis=0)
            wpos_ids = wpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            wpos_ids = mx.transpose(wpos_ids, (0, 2, 1, 3))
            wpos_ids = wpos_ids.flatten()

            stacked_pos_ids = mx.stack([hpos_ids, wpos_ids], axis=-1)
            pos_ids.append(mx.tile(stacked_pos_ids, (t, 1)))

        pos_ids = mx.concatenate(pos_ids, axis=0)
        max_grid_size = mx.max(grid_thw[:, 1:])
        rotary_pos_emb_full = self.rotary_pos_emb(max_grid_size.item())
        rotary_pos_emb = rotary_pos_emb_full[pos_ids].reshape(pos_ids.shape[0], -1)

        emb = mx.concatenate((rotary_pos_emb, rotary_pos_emb), axis=-1)
        return (mx.cos(emb), mx.sin(emb)), pos_ids

    def __call__(
        self,
        hidden_states: mx.array,
        grid_thw: mx.array,
        output_hidden_states: Optional[bool] = None,
    ) -> mx.array:
        hidden_states = self.patch_embed(hidden_states)
        position_embeddings, _ = self.rot_pos_emb(grid_thw)

        seq_lens = grid_thw[:, 1] * grid_thw[:, 2]
        repeats = grid_thw[:, 0]
        repeated_values = []
        for seq_len, repeat_count in zip(seq_lens.tolist(), repeats.tolist()):
            repeated_values.extend([seq_len] * repeat_count)

        cu_seqlens = mx.array(repeated_values).cumsum(axis=0)
        cu_seqlens = mx.pad(cu_seqlens, (1, 0), constant_values=0)

        for blk in self.blocks:
            hidden_states = blk(
                hidden_states,
                cu_seqlens=cu_seqlens,
                position_embeddings=position_embeddings,
            )

        hidden_states = self.post_layernorm(hidden_states)

        hidden_states = hidden_states.reshape(
            -1,
            self.spatial_merge_size,
            self.spatial_merge_size,
            hidden_states.shape[-1],
        )
        hidden_states = self.downsample(hidden_states).reshape(
            -1, self.config.out_hidden_size
        )

        merged_hidden_states = self.merger(hidden_states)
        return merged_hidden_states

    def sanitize(self, weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if "position_ids" in k:
                continue
            elif "patch_embed.proj.weight" in k or "downsample.weight" in k:
                if check_array_shape(v):
                    sanitized_weights[k] = v
                else:
                    if v.ndim == 5:
                        sanitized_weights[k] = v.transpose(0, 2, 3, 4, 1)
                    elif v.ndim == 4:
                        sanitized_weights[k] = v.transpose(0, 2, 3, 1)
                    else:
                        sanitized_weights[k] = v
            else:
                sanitized_weights[k] = v

        return sanitized_weights



from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from mlx_vlm.models.qwen3_omni_moe.audio import AudioModel

from ..base import InputEmbeddingsFeatures
from .config import ThinkerConfig
from .language import LanguageModel
from .vision import VisionModel


def masked_scatter(
    final_embedding: mx.array,
    image_mask_expanded: mx.array,
    scaled_image_features: mx.array,
):
    final_embedding_shape = final_embedding.shape
    scaled_image_features_flattened = mx.flatten(scaled_image_features)
    final_embedding_flattened = mx.flatten(final_embedding)
    image_mask_expanded_flattened = mx.flatten(image_mask_expanded)

    image_positions = mx.array(np.where(image_mask_expanded_flattened)[0], mx.uint32)
    final_embedding_flattened[image_positions] = scaled_image_features_flattened

    final_embedding = mx.reshape(final_embedding_flattened, final_embedding_shape)

    return final_embedding


class Thinker(nn.Module):
    def __init__(self, config: ThinkerConfig):
        super().__init__()
        self.config = config

        self.vision_tower = VisionModel(config.vision_config)

        self.audio_tower = AudioModel(config.audio_config)

        self.language_model = LanguageModel(config.text_config, config)

    def get_audio_features(
        self,
        input_features: mx.array,
        feature_attention_mask: Optional[mx.array] = None,
        audio_feature_lengths: Optional[mx.array] = None,
    ):
        if feature_attention_mask is not None:
            input_features = mx.transpose(input_features, (0, 2, 1))
            audio_feature_lengths = mx.sum(feature_attention_mask, axis=1)
            mask_bool = feature_attention_mask.astype(mx.bool_)
            batch_size, seq_len, hidden_dim = input_features.shape
            input_features_flat = mx.reshape(input_features, (-1, hidden_dim))
            mask_flat = mx.reshape(mask_bool, (-1,))
            indices = mx.array(np.where(mask_flat)[0])
            input_features = mx.take(input_features_flat, indices, axis=0)
            input_features = mx.transpose(input_features, (1, 0))

        feature_lens = (
            audio_feature_lengths
            if audio_feature_lengths is not None
            else (
                mx.sum(feature_attention_mask, axis=-1)
                if feature_attention_mask is not None
                else None
            )
        )
        audio_outputs = self.audio_tower(
            input_features,
            feature_lens=feature_lens,
        )
        return audio_outputs

    def get_placeholder_mask(
        self,
        input_ids: Optional[mx.array] = None,
        inputs_embeds: Optional[mx.array] = None,
        image_features: Optional[mx.array] = None,
        video_features: Optional[mx.array] = None,
        audio_features: Optional[mx.array] = None,
    ):
        if input_ids is None:
            image_token_embed = self.language_model.model.embed_tokens(
                mx.array([self.config.image_token_id], dtype=mx.int32)
            )
            video_token_embed = self.language_model.model.embed_tokens(
                mx.array([self.config.video_token_id], dtype=mx.int32)
            )
            audio_token_embed = self.language_model.model.embed_tokens(
                mx.array([self.config.audio_token_id], dtype=mx.int32)
            )

            special_image_mask = mx.all(inputs_embeds == image_token_embed, axis=-1)
            special_video_mask = mx.all(inputs_embeds == video_token_embed, axis=-1)
            special_audio_mask = mx.all(inputs_embeds == audio_token_embed, axis=-1)
        else:
            special_image_mask = input_ids == self.config.image_token_id
            special_video_mask = input_ids == self.config.video_token_id
            special_audio_mask = input_ids == self.config.audio_token_id

        n_image_tokens = mx.sum(special_image_mask)
        special_image_mask = mx.expand_dims(special_image_mask, axis=-1)
        special_image_mask = mx.broadcast_to(special_image_mask, inputs_embeds.shape)
        if (
            image_features is not None
            and mx.sum(special_image_mask) != image_features.size
        ):
            raise ValueError(
                f"Image features and image tokens do not match: tokens: {n_image_tokens}, features {image_features.shape[0]}"
            )

        n_video_tokens = mx.sum(special_video_mask)
        special_video_mask = mx.expand_dims(special_video_mask, axis=-1)
        special_video_mask = mx.broadcast_to(special_video_mask, inputs_embeds.shape)
        if (
            video_features is not None
            and mx.sum(special_video_mask) != video_features.size
        ):
            raise ValueError(
                f"Video features and video tokens do not match: tokens: {n_video_tokens}, features {video_features.shape[0]}"
            )

        special_audio_mask = mx.expand_dims(special_audio_mask, axis=-1)
        special_audio_mask = mx.broadcast_to(special_audio_mask, inputs_embeds.shape)
        return special_image_mask, special_video_mask, special_audio_mask

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        pixel_values_videos: Optional[mx.array] = None,
        image_grid_thw: Optional[mx.array] = None,
        video_grid_thw: Optional[mx.array] = None,
        input_features: Optional[mx.array] = None,
        feature_attention_mask: Optional[mx.array] = None,
        audio_feature_lengths: Optional[mx.array] = None,
        **kwargs,
    ):
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)
        visual_pos_masks = None
        deepstack_visual_embeds = None

        if input_features is not None:
            audio_features = self.get_audio_features(
                input_features,
                feature_attention_mask=feature_attention_mask,
                audio_feature_lengths=audio_feature_lengths,
            )
            audio_features = audio_features.astype(inputs_embeds.dtype)
            _, _, audio_mask = self.get_placeholder_mask(
                input_ids, inputs_embeds=inputs_embeds
            )
            if audio_features.ndim == 2:
                audio_mask_flat = mx.reshape(
                    audio_mask[..., 0] if audio_mask.ndim == 3 else audio_mask, (-1,)
                )
                num_audio_tokens = int(mx.sum(audio_mask_flat))
                if num_audio_tokens > 0:
                    if audio_features.shape[0] != num_audio_tokens:
                        if audio_features.shape[0] > num_audio_tokens:
                            audio_features = audio_features[:num_audio_tokens]
                        else:
                            padding = mx.zeros(
                                (
                                    num_audio_tokens - audio_features.shape[0],
                                    audio_features.shape[1],
                                ),
                                dtype=audio_features.dtype,
                            )
                            audio_features = mx.concatenate(
                                [audio_features, padding], axis=0
                            )
            inputs_embeds = masked_scatter(inputs_embeds, audio_mask, audio_features)

        if pixel_values is not None:
            dtype = self.vision_tower.patch_embed.proj.weight.dtype
            pixel_values = pixel_values.astype(dtype)
            vision_output = self.vision_tower(pixel_values, image_grid_thw)
            if isinstance(vision_output, tuple):
                image_embeds, image_embeds_multiscale = vision_output
            else:
                image_embeds = vision_output
                image_embeds_multiscale = None
            image_embeds = image_embeds.astype(inputs_embeds.dtype)
            image_mask, _, _ = self.get_placeholder_mask(
                input_ids, inputs_embeds=inputs_embeds, image_features=image_embeds
            )
            inputs_embeds = masked_scatter(inputs_embeds, image_mask, image_embeds)
            visual_pos_masks = (
                image_mask[..., 0] if image_mask.ndim == 3 else image_mask
            )
            visual_embeds_multiscale = image_embeds_multiscale

        if pixel_values_videos is not None:
            dtype = self.vision_tower.patch_embed.proj.weight.dtype
            pixel_values_videos = pixel_values_videos.astype(dtype)
            vision_output = self.vision_tower(pixel_values_videos, video_grid_thw)
            if isinstance(vision_output, tuple):
                video_embeds, video_embeds_multiscale = vision_output
            else:
                video_embeds = vision_output
                video_embeds_multiscale = None
            video_embeds = video_embeds.astype(inputs_embeds.dtype)
            _, video_mask, _ = self.get_placeholder_mask(
                input_ids, inputs_embeds=inputs_embeds, video_features=video_embeds
            )
            inputs_embeds = masked_scatter(inputs_embeds, video_mask, video_embeds)

            if visual_embeds_multiscale is None:
                visual_embeds_multiscale = video_embeds_multiscale
                visual_pos_masks = (
                    video_mask[..., 0] if video_mask.ndim == 3 else video_mask
                )
            else:
                visual_pos_masks = (
                    (video_mask | image_mask)[..., 0]
                    if video_mask.ndim == 3
                    else (video_mask | image_mask)
                )
                visual_mask_flat = mx.reshape(visual_pos_masks, (-1,))
                image_mask_flat = mx.reshape(
                    image_mask[..., 0] if image_mask.ndim == 3 else image_mask, (-1,)
                )
                video_mask_flat = mx.reshape(
                    video_mask[..., 0] if video_mask.ndim == 3 else video_mask, (-1,)
                )
                visual_indices_flat = mx.where(visual_mask_flat)[0]
                image_mask_on_visual = mx.take(
                    image_mask_flat, visual_indices_flat, axis=0
                )
                video_mask_on_visual = mx.take(
                    video_mask_flat, visual_indices_flat, axis=0
                )
                image_indices = mx.where(image_mask_on_visual)[0]
                video_indices = mx.where(video_mask_on_visual)[0]

                visual_embeds_multiscale_joint = []
                for img_embed, vid_embed in zip(
                    visual_embeds_multiscale, video_embeds_multiscale
                ):
                    embed_joint = mx.zeros(
                        (len(visual_indices_flat), img_embed.shape[-1]),
                        dtype=img_embed.dtype,
                    )
                    if len(image_indices) > 0:
                        embed_joint = mx.scatter(
                            embed_joint,
                            image_indices,
                            mx.take(img_embed, image_indices, axis=0),
                            axis=0,
                        )
                    if len(video_indices) > 0:
                        embed_joint = mx.scatter(
                            embed_joint,
                            video_indices,
                            mx.take(vid_embed, video_indices, axis=0),
                            axis=0,
                        )
                    visual_embeds_multiscale_joint.append(embed_joint)
                visual_embeds_multiscale = tuple(visual_embeds_multiscale_joint)

        return InputEmbeddingsFeatures(
            inputs_embeds=inputs_embeds,
            visual_pos_masks=visual_pos_masks,
            deepstack_visual_embeds=deepstack_visual_embeds,
        )

    @staticmethod
    def merge_input_ids_with_image_features(
        image_features, inputs_embeds, input_ids, image_token_index, video_token_index
    ):
        special_image_mask = input_ids == image_token_index
        special_video_mask = input_ids == video_token_index
        special_image_mask = special_image_mask | special_video_mask
        n_image_tokens = special_image_mask.sum()
        special_image_mask = special_image_mask[..., None]
        special_image_mask = mx.broadcast_to(special_image_mask, inputs_embeds.shape)

        n_image_features = image_features.shape[0]
        n_image_mask_elements = special_image_mask.sum()
        if n_image_mask_elements != image_features.size:
            raise ValueError(
                f"Image features and image tokens do not match: tokens: {n_image_tokens}, features {n_image_features}"
            )

        inputs_embeds = masked_scatter(
            inputs_embeds, special_image_mask, image_features
        )

        return inputs_embeds, special_image_mask

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        pixel_values_videos: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        image_grid_thw = kwargs.pop("image_grid_thw", None)
        video_grid_thw = kwargs.pop("video_grid_thw", None)
        input_features = kwargs.pop("input_features", None)
        feature_attention_mask = kwargs.pop("feature_attention_mask", None)
        audio_feature_lengths = kwargs.pop("audio_feature_lengths", None)

        inputs_embeds, visual_pos_masks, deepstack_visual_embeds = (
            self.get_input_embeddings(
                input_ids,
                pixel_values=pixel_values,
                pixel_values_videos=pixel_values_videos,
                image_grid_thw=image_grid_thw,
                video_grid_thw=video_grid_thw,
                input_features=input_features,
                feature_attention_mask=feature_attention_mask,
                audio_feature_lengths=audio_feature_lengths,
            )
        )

        kwargs = {
            "pixel_values": pixel_values,
            "pixel_values_videos": pixel_values_videos,
            "image_grid_thw": image_grid_thw,
            "video_grid_thw": video_grid_thw,
            "visual_pos_masks": visual_pos_masks,
            "deepstack_visual_embeds": deepstack_visual_embeds,
            **kwargs,
        }

        logits = self.language_model(
            input_ids, inputs_embeds, mask=mask, cache=cache, **kwargs
        )

        return logits

    def sanitize(self, weights):
        sanitized_weights = {}
        for key, value in weights.items():
            if "thinker" in key:
                if "thinker.model" in key:
                    key = key.replace("thinker.model", "thinker.language_model.model")
                elif "thinker.visual" in key:
                    key = key.replace("thinker.visual", "thinker.vision_tower")
                    if "merger" in key:
                        key = (
                            key.replace("ln_q", "norm")
                            .replace("mlp.0", "linear_fc1")
                            .replace("mlp.2", "linear_fc2")
                        )
                    if "merger_list" in key:
                        key = key.replace("merger_list", "deepstack_merger_list")

                if "thinker.lm_head" in key:
                    key = key.replace(
                        "thinker.lm_head", "thinker.language_model.lm_head"
                    )

            sanitized_weights[key] = value

        return sanitized_weights



from .config import AdapterConfig, ModelConfig, TextConfig, VisionConfig, VitConfig
from .language import LanguageModel
from .molmo2 import Model
from .processing import Molmo2ImageProcessor as ImageProcessor
from .processing import Molmo2Processor as Processor
from .vision import VisionModel



"""Language model for ERNIE 4.5 VL MoE."""

from typing import Optional

import mlx.core as mx
import mlx.nn as nn
from mlx_lm.models.switch_layers import SwitchGLU

from ..base import (
    LanguageModelOutput,
    create_attention_mask,
    scaled_dot_product_attention,
)
from ..cache import KVCache
from .config import ModelConfig, TextConfig


class Ernie4_5RotaryEmbedding:
    """Rotary Position Embedding for ERNIE 4.5 VL with MRoPE support.

    Matches PyTorch's implementation with pre-rotated inverse frequencies.
    """

    def __init__(
        self,
        dim: int,
        base: float = 10000,
        mrope_section: tuple = (22, 22, 20),
    ):
        self.dim = dim  # head_dim
        self.base = base
        self.mrope_section = mrope_section  # (h_dim, w_dim, t_dim)

        # Pre-compute inverse frequencies
        indices = mx.arange(0, self.dim, 2, dtype=mx.float32)
        inv_freq = 1.0 / (self.base ** (indices / self.dim))

        # Pre-rotate frequencies to match PyTorch's approach
        # This avoids rotation during forward pass
        hw_dim = mrope_section[0] + mrope_section[1]  # 44
        t_dim = mrope_section[2]  # 20

        inv_freq_3d = mx.zeros_like(inv_freq)
        # Pre-rotate HW dimensions: [even, odd] -> interleaved during recomposition
        hw_freqs = inv_freq[:-t_dim]  # First (dim/2 - t_dim) frequencies
        inv_freq_3d = mx.concatenate(
            [
                mx.concatenate([hw_freqs[0::2], hw_freqs[1::2]]),  # Pre-rotated HW
                inv_freq[-t_dim:],  # T frequencies unchanged
            ]
        )
        self.inv_freq = inv_freq_3d

    def _recomposition_to_3d(self, freq):
        """Recompose frequencies for 3D positions matching PyTorch's approach.

        Args:
            freq: [3, batch, seq_len, dim//2] - frequencies for T, H, W dimensions

        Returns:
            Recomposed frequencies [batch, seq_len, dim]
        """
        # Split by mrope_section
        h_dim, w_dim, t_dim = self.mrope_section

        # freq shape: [3, batch, seq_len, half_dim]
        # Split each dimension's frequencies
        freq_parts = []
        for i in range(3):
            freq_parts.append(mx.split(freq[i], [h_dim, h_dim + w_dim], axis=-1))

        # Recompose: freq_h from dim 1, freq_w from dim 2, freq_t from dim 0
        # This matches PyTorch's (i + 1) % 3 indexing
        freq_h = freq_parts[1][0]  # H from position 1
        freq_w = freq_parts[2][1]  # W from position 2
        freq_t = freq_parts[0][2]  # T from position 0

        # Interleave H and W: [h0, w0, h1, w1, ...]
        freq_hw = mx.stack([freq_h, freq_w], axis=-1).reshape(
            freq_h.shape[0], freq_h.shape[1], -1
        )

        # Concatenate HW and T
        freq_hwt = mx.concatenate([freq_hw, freq_t], axis=-1)

        # Repeat interleave by 2 for full head_dim
        freq_full = mx.repeat(freq_hwt, 2, axis=-1)

        return freq_full

    def __call__(self, x, position_ids):
        """
        Compute 3D rotary embeddings matching PyTorch's implementation.

        Args:
            x: Input tensor for dtype reference
            position_ids: Position IDs, shape (batch, seq_len, 3) for 3D positions [T, H, W]

        Returns:
            cos, sin: [batch, seq_len, head_dim] ready for rotation
        """
        if position_ids.ndim == 2:
            # 1D positions - expand to 3D with same values
            position_ids = mx.stack([position_ids, position_ids, position_ids], axis=-1)

        batch_size, seq_len, _ = position_ids.shape

        # position_ids: [batch, seq_len, 3] -> [3, batch, seq_len]
        position_ids = position_ids.transpose(2, 0, 1).astype(mx.float32)

        # inv_freq: [dim//2] -> [1, 1, dim//2, 1] for broadcasting
        inv_freq_expanded = self.inv_freq[None, None, :, None]  # [1, 1, dim//2, 1]
        inv_freq_expanded = mx.broadcast_to(
            inv_freq_expanded, (3, batch_size, self.dim // 2, 1)
        )

        # position_ids: [3, batch, seq_len] -> [3, batch, 1, seq_len]
        position_ids_expanded = position_ids[:, :, None, :]

        # freqs: [3, batch, dim//2, seq_len] -> [3, batch, seq_len, dim//2]
        freqs = (inv_freq_expanded * position_ids_expanded).transpose(0, 1, 3, 2)

        cos = mx.cos(freqs)
        sin = mx.sin(freqs)

        # Recompose to 3D
        cos = self._recomposition_to_3d(cos)
        sin = self._recomposition_to_3d(sin)

        return cos.astype(x.dtype), sin.astype(x.dtype)


def rotate_half_interleaved(x):
    """Rotates using interleaved pattern: [-x1, x0, -x3, x2, ...].

    This matches PyTorch's rotation: stack([-x[1::2], x[0::2]], dim=-1).reshape()
    """
    x_even = x[..., 0::2]  # [x0, x2, x4, ...]
    x_odd = x[..., 1::2]  # [x1, x3, x5, ...]
    # Stack as [-odd, even] and reshape
    rotated = mx.stack([-x_odd, x_even], axis=-1)
    return rotated.reshape(x.shape)


def apply_rotary_pos_emb(q, k, cos_pos, sin_pos):
    """Apply rotary position embeddings to queries and keys.

    Uses interleaved rotation matching PyTorch's apply_rotary_3d.

    Args:
        q: [batch, n_heads, seq_len, head_dim]
        k: [batch, n_kv_heads, seq_len, head_dim]
        cos_pos: [batch, seq_len, head_dim]
        sin_pos: [batch, seq_len, head_dim]
    """
    orig_dtype = q.dtype
    # Expand for heads dimension

    cos_pos = mx.expand_dims(cos_pos, axis=1)  # [batch, 1, seq_len, head_dim]
    sin_pos = mx.expand_dims(sin_pos, axis=1)

    # Apply rotation: q_rotated = q * cos + rotate_half(q) * sin
    q_rotated = rotate_half_interleaved(q)
    k_rotated = rotate_half_interleaved(k)

    q_embed = (q.astype(mx.float32) * cos_pos) + (
        q_rotated.astype(mx.float32) * sin_pos
    )
    k_embed = (k.astype(mx.float32) * cos_pos) + (
        k_rotated.astype(mx.float32) * sin_pos
    )

    return q_embed.astype(orig_dtype), k_embed.astype(orig_dtype)


class Attention(nn.Module):
    """Multi-headed attention for ERNIE 4.5 with MRoPE support."""

    def __init__(self, args: TextConfig):
        super().__init__()

        dim = args.hidden_size
        self.n_heads = n_heads = args.num_attention_heads
        self.n_kv_heads = n_kv_heads = args.num_key_value_heads or n_heads

        self.head_dim = head_dim = args.hidden_size // n_heads
        self.scale = head_dim**-0.5

        self.q_proj = nn.Linear(dim, n_heads * head_dim, bias=args.use_bias)
        self.k_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=args.use_bias)
        self.v_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=args.use_bias)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=args.use_bias)

        # Get mrope_section for 3D RoPE (H, W, T dimension allocation)
        # Default [22, 22, 20] for head_dim=128
        self.mrope_section = tuple(getattr(args, "mrope_section", [22, 22, 20]))

        self.rotary_emb = Ernie4_5RotaryEmbedding(
            head_dim,
            base=args.rope_theta,
            mrope_section=self.mrope_section,
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
        position_ids: Optional[mx.array] = None,
    ) -> mx.array:
        B, L, D = x.shape

        queries = self.q_proj(x)
        keys = self.k_proj(x)
        values = self.v_proj(x)

        # Reshape and transpose: [B, L, n_heads, head_dim] -> [B, n_heads, L, head_dim]
        queries = queries.reshape(B, L, self.n_heads, self.head_dim).transpose(
            0, 2, 1, 3
        )
        keys = keys.reshape(B, L, self.n_kv_heads, self.head_dim).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, self.head_dim).transpose(
            0, 2, 1, 3
        )

        # Handle position IDs
        if position_ids is None:
            offset = cache.offset if cache is not None else 0
            position_ids = mx.arange(offset, offset + L)
            position_ids = mx.expand_dims(position_ids, axis=0)

        cos, sin = self.rotary_emb(values, position_ids)
        queries, keys = apply_rotary_pos_emb(queries, keys, cos, sin)

        if cache is not None:
            keys, values = cache.update_and_fetch(keys, values)

        if mask is not None and isinstance(mask, mx.array):
            mask = mask[..., : keys.shape[-2]]

        output = scaled_dot_product_attention(
            queries, keys, values, cache, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output)


class Ernie4_5_MLP(nn.Module):
    def __init__(self, dim, hidden_dim, use_bias=False):
        super().__init__()
        self.gate_proj = nn.Linear(dim, hidden_dim, bias=use_bias)
        self.down_proj = nn.Linear(hidden_dim, dim, bias=use_bias)
        self.up_proj = nn.Linear(dim, hidden_dim, bias=use_bias)

    def __call__(self, x) -> mx.array:
        return self.down_proj(nn.silu(self.gate_proj(x)) * self.up_proj(x))


class Ernie4_5_MoeMLP(nn.Module):
    """Mixture of Experts MLP for ERNIE with dual expert groups."""

    def __init__(self, args: TextConfig):
        super().__init__()
        self.args = args
        self.k = args.moe_k
        self.norm_min = getattr(args, "moe_norm_min", 1e-12)

        moe_num_experts = args.moe_num_experts
        moe_intermediate_size = args.moe_intermediate_size

        if isinstance(moe_num_experts, (list, tuple)) and len(moe_num_experts) == 2:
            self.num_text_experts = moe_num_experts[0]
            self.num_mm_experts = moe_num_experts[1]
            self.has_dual_experts = True
        else:
            self.num_text_experts = (
                moe_num_experts
                if not isinstance(moe_num_experts, (list, tuple))
                else moe_num_experts[0]
            )
            self.num_mm_experts = 0
            self.has_dual_experts = False

        if (
            isinstance(moe_intermediate_size, (list, tuple))
            and len(moe_intermediate_size) == 2
        ):
            self.text_intermediate_size = moe_intermediate_size[0]
            self.mm_intermediate_size = moe_intermediate_size[1]
        else:
            self.text_intermediate_size = (
                moe_intermediate_size
                if not isinstance(moe_intermediate_size, (list, tuple))
                else moe_intermediate_size[0]
            )
            self.mm_intermediate_size = self.text_intermediate_size

        self.gate = nn.Linear(args.hidden_size, self.num_text_experts, bias=False)
        self.e_score_correction_bias = mx.zeros((self.num_text_experts,))
        self.switch_mlp = SwitchGLU(
            args.hidden_size,
            self.text_intermediate_size,
            self.num_text_experts,
            bias=args.use_bias,
        )

        if self.has_dual_experts and self.num_mm_experts > 0:
            self.gate_1 = nn.Linear(args.hidden_size, self.num_mm_experts, bias=False)
            self.e_score_correction_bias_1 = mx.zeros((self.num_mm_experts,))
            self.switch_mlp_1 = SwitchGLU(
                args.hidden_size,
                self.mm_intermediate_size,
                self.num_mm_experts,
                bias=args.use_bias,
            )

        if getattr(args, "moe_num_shared_experts", 0) > 0:
            shared_intermediate_size = (
                self.text_intermediate_size * args.moe_num_shared_experts
            )
            self.shared_experts = Ernie4_5_MLP(
                args.hidden_size, shared_intermediate_size, args.use_bias
            )
        else:
            self.shared_experts = None

    def _route_experts(
        self, x: mx.array, gate: nn.Module, e_score_correction_bias: mx.array
    ) -> tuple:
        k = self.k
        router_logits = gate(x).astype(mx.float32)
        routing_weights = mx.softmax(router_logits, axis=-1)
        routing_weights_with_bias = routing_weights + e_score_correction_bias

        selected_experts = mx.stop_gradient(
            mx.argpartition(-routing_weights_with_bias, kth=k - 1, axis=-1)[..., :k]
        )
        scores = mx.take_along_axis(routing_weights, selected_experts, axis=-1)
        scores = scores / mx.maximum(scores.sum(axis=-1, keepdims=True), self.norm_min)

        return selected_experts, scores

    def __call__(
        self, x: mx.array, token_type_ids: Optional[mx.array] = None
    ) -> mx.array:
        inds, scores = self._route_experts(x, self.gate, self.e_score_correction_bias)
        y_text = self.switch_mlp(x, inds)
        y_text = (y_text * scores[..., None]).sum(axis=-2).astype(y_text.dtype)

        if (
            not self.has_dual_experts
            or self.num_mm_experts == 0
            or token_type_ids is None
        ):
            y = y_text
        else:
            inds_mm, scores_mm = self._route_experts(
                x, self.gate_1, self.e_score_correction_bias_1
            )
            y_mm = self.switch_mlp_1(x, inds_mm)
            y_mm = (y_mm * scores_mm[..., None]).sum(axis=-2).astype(y_mm.dtype)

            is_text = token_type_ids == 0
            is_text_expanded = mx.expand_dims(is_text, axis=-1)
            y = mx.where(is_text_expanded, y_text, y_mm)

        if self.shared_experts is not None:
            y = y + self.shared_experts(x)

        return y


class Ernie4_5VLDecoderLayer(nn.Module):
    def __init__(self, args: TextConfig, layer_idx: int):
        super().__init__()
        self.hidden_size = args.hidden_size
        self.self_attn = Attention(args)

        moe_layer_start_index = args.moe_layer_start_index
        if isinstance(moe_layer_start_index, (tuple, list)):
            moe_layer_start_index = min(moe_layer_start_index)

        moe_layer_end_index = args.moe_layer_end_index
        if moe_layer_end_index is None:
            moe_layer_end_index = args.num_hidden_layers - 1
        elif isinstance(moe_layer_end_index, (tuple, list)):
            moe_layer_end_index = max(moe_layer_end_index)

        use_moe = (
            ((layer_idx + 1) % args.moe_layer_interval == 0)
            and layer_idx >= moe_layer_start_index
            and layer_idx <= moe_layer_end_index
        )

        if use_moe:
            self.mlp = Ernie4_5_MoeMLP(args)
        else:
            self.mlp = Ernie4_5_MLP(
                args.hidden_size, args.intermediate_size, args.use_bias
            )

        self.input_layernorm = nn.RMSNorm(args.hidden_size, eps=args.rms_norm_eps)
        self.post_attention_layernorm = nn.RMSNorm(
            args.hidden_size, eps=args.rms_norm_eps
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[KVCache] = None,
        position_ids: Optional[mx.array] = None,
        token_type_ids: Optional[mx.array] = None,
    ) -> mx.array:
        r = self.self_attn(self.input_layernorm(x), mask, cache, position_ids)
        h = x + r
        if isinstance(self.mlp, Ernie4_5_MoeMLP):
            r = self.mlp(
                self.post_attention_layernorm(h), token_type_ids=token_type_ids
            )
        else:
            r = self.mlp(self.post_attention_layernorm(h))
        return h + r


class Ernie4_5Model(nn.Module):
    def __init__(self, args: TextConfig):
        super().__init__()
        self.args = args
        self.vocab_size = args.vocab_size
        self.num_hidden_layers = args.num_hidden_layers

        self.embed_tokens = nn.Embedding(args.vocab_size, args.hidden_size)
        self.layers = [
            Ernie4_5VLDecoderLayer(args=args, layer_idx=i)
            for i in range(args.num_hidden_layers)
        ]
        self.norm = nn.RMSNorm(args.hidden_size, eps=args.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        position_ids: Optional[mx.array] = None,
        token_type_ids: Optional[mx.array] = None,
    ):
        if inputs_embeds is None:
            h = self.embed_tokens(inputs)
        else:
            h = inputs_embeds

        if cache is None:
            cache = [None] * len(self.layers)

        if mask is None:
            mask = create_attention_mask(h, cache)

        for layer, c in zip(self.layers, cache):
            h = layer(h, mask, c, position_ids, token_type_ids=token_type_ids)

        return self.norm(h)


class LanguageModel(nn.Module):
    def __init__(self, args: TextConfig, config: ModelConfig = None):
        super().__init__()
        self.args = args
        self.config = config
        self.model_type = args.model_type
        self.model = Ernie4_5Model(args)
        self._rope_deltas = None

        if not args.tie_word_embeddings:
            self.lm_head = nn.Linear(args.hidden_size, args.vocab_size, bias=False)

    def get_rope_index(
        self,
        input_ids: mx.array,
        image_grid_thw: Optional[mx.array] = None,
        video_grid_thw: Optional[mx.array] = None,
        attention_mask: Optional[mx.array] = None,
    ):
        batch_size, seq_length = input_ids.shape
        spatial_merge_size = self.config.vision_config.spatial_merge_size
        image_token_id = self.config.image_token_id
        video_token_id = self.config.video_token_id
        vision_start_token_id = self.config.vision_start_token_id

        if image_grid_thw is not None or video_grid_thw is not None:
            batch_position_ids = []
            mrope_position_deltas = []

            image_index, video_index = 0, 0

            for i in range(batch_size):
                input_tokens = input_ids[i].tolist()
                llm_pos_ids_list = []
                st = 0

                image_nums, video_nums = 0, 0
                for idx, token in enumerate(input_tokens):
                    if token == vision_start_token_id and idx + 1 < len(input_tokens):
                        next_token = input_tokens[idx + 1]
                        if next_token == image_token_id:
                            image_nums += 1
                        elif next_token == video_token_id:
                            video_nums += 1

                remain_images, remain_videos = image_nums, video_nums

                for _ in range(image_nums + video_nums):
                    ed_image = (
                        input_tokens.index(image_token_id, st)
                        if image_token_id in input_tokens[st:] and remain_images > 0
                        else len(input_tokens) + 1
                    )
                    ed_video = (
                        input_tokens.index(video_token_id, st)
                        if video_token_id in input_tokens[st:] and remain_videos > 0
                        else len(input_tokens) + 1
                    )

                    if ed_image < ed_video:
                        t, h, w = image_grid_thw[image_index].tolist()
                        image_index += 1
                        remain_images -= 1
                        ed = ed_image
                        vision_token = image_token_id
                    else:
                        t, h, w = video_grid_thw[video_index].tolist()
                        video_index += 1
                        remain_videos -= 1
                        ed = ed_video
                        vision_token = video_token_id

                    llm_grid_t = t
                    llm_grid_h = h // spatial_merge_size
                    llm_grid_w = w // spatial_merge_size
                    expected_vision_len = llm_grid_t * llm_grid_h * llm_grid_w

                    actual_vision_len = 0
                    for j in range(
                        ed, min(ed + expected_vision_len, len(input_tokens))
                    ):
                        if input_tokens[j] == vision_token:
                            actual_vision_len += 1
                        else:
                            break

                    text_len = ed - st
                    st_idx = llm_pos_ids_list[-1].max() + 1 if llm_pos_ids_list else 0

                    text_pos = mx.arange(text_len) + st_idx
                    text_pos_3d = mx.stack([text_pos, text_pos, text_pos], axis=0)
                    llm_pos_ids_list.append(text_pos_3d)

                    if actual_vision_len > 0:
                        t_idx = mx.repeat(
                            mx.arange(llm_grid_t).reshape(-1, 1),
                            llm_grid_h * llm_grid_w,
                            axis=1,
                        ).flatten()[:actual_vision_len]
                        h_idx = mx.tile(
                            mx.arange(llm_grid_h).reshape(1, -1, 1),
                            (llm_grid_t, 1, llm_grid_w),
                        ).flatten()[:actual_vision_len]
                        w_idx = mx.tile(
                            mx.arange(llm_grid_w).reshape(1, 1, -1),
                            (llm_grid_t, llm_grid_h, 1),
                        ).flatten()[:actual_vision_len]

                        vision_pos = (
                            mx.stack([t_idx, h_idx, w_idx], axis=0) + text_len + st_idx
                        )
                        llm_pos_ids_list.append(vision_pos)

                    st = ed + actual_vision_len

                # Handle remaining text
                if st < len(input_tokens):
                    st_idx = llm_pos_ids_list[-1].max() + 1 if llm_pos_ids_list else 0
                    text_len = len(input_tokens) - st
                    text_pos = mx.arange(text_len) + st_idx
                    text_pos_3d = mx.stack([text_pos, text_pos, text_pos], axis=0)
                    llm_pos_ids_list.append(text_pos_3d)

                llm_positions = mx.concatenate(llm_pos_ids_list, axis=1)  # [3, seq_len]
                batch_position_ids.append(llm_positions.T)  # [seq_len, 3]
                mrope_position_deltas.append(llm_positions.max() + 1 - seq_length)

            position_ids = mx.stack(batch_position_ids, axis=0)
            mrope_position_deltas = mx.array(mrope_position_deltas)
            return position_ids, mrope_position_deltas
        else:
            position_ids = mx.arange(seq_length)
            position_ids = mx.broadcast_to(
                position_ids[None, :], (batch_size, seq_length)
            )
            position_ids = mx.stack([position_ids, position_ids, position_ids], axis=-1)
            return position_ids, mx.zeros((batch_size,), dtype=mx.int32)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):
        position_ids = kwargs.pop("position_ids", None)
        pixel_values = kwargs.pop("pixel_values", None)
        image_grid_thw = kwargs.pop("image_grid_thw", None)
        video_grid_thw = kwargs.pop("video_grid_thw", None)

        if pixel_values is not None:
            self._rope_deltas = None

        cache_offset = 0
        if cache and cache[0] is not None:
            offset = cache[0].offset
            cache_offset = offset.item() if isinstance(offset, mx.array) else offset

        if position_ids is None and (mask is None or mask.ndim == 2):
            if (
                cache is None or cache[0] is None or cache_offset == 0
            ) or self._rope_deltas is None:
                position_ids, rope_deltas = self.get_rope_index(
                    inputs, image_grid_thw, video_grid_thw, mask
                )
                self._rope_deltas = rope_deltas
            else:
                batch_size, seq_length = inputs.shape
                delta = cache_offset + self._rope_deltas if cache is not None else 0
                position_ids = mx.arange(seq_length) + delta
                position_ids = mx.broadcast_to(
                    position_ids[None, :], (batch_size, seq_length)
                )
                position_ids = mx.stack(
                    [position_ids, position_ids, position_ids], axis=-1
                )

        token_type_ids = kwargs.pop("token_type_ids", None)

        out = self.model(
            inputs,
            cache=cache,
            inputs_embeds=inputs_embeds,
            mask=mask,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
        )

        if self.args.tie_word_embeddings:
            out = self.model.embed_tokens.as_linear(out)
        else:
            out = self.lm_head(out)

        return LanguageModelOutput(logits=out)

    @property
    def layers(self):
        return self.model.layers

    @property
    def head_dim(self):
        return self.args.hidden_size // self.args.num_attention_heads

    @property
    def n_kv_heads(self):
        return self.args.num_key_value_heads

    def sanitize(self, weights):
        """Sanitize weights for loading."""
        remove_patterns = [
            "mtp_block.",
            "mtp_linear_proj.",
            "mtp_hidden_norm.",
            "mtp_emb_norm.",
        ]

        weights = {
            key: value
            for key, value in weights.items()
            if not any(pattern in key for pattern in remove_patterns)
        }

        # Get expert configuration
        moe_num_experts = self.args.moe_num_experts
        if isinstance(moe_num_experts, (list, tuple)) and len(moe_num_experts) == 2:
            num_text_experts = moe_num_experts[0]
            num_mm_experts = moe_num_experts[1]
        else:
            num_text_experts = (
                moe_num_experts
                if not isinstance(moe_num_experts, (list, tuple))
                else moe_num_experts[0]
            )
            num_mm_experts = 0

        for l in range(self.args.num_hidden_layers):
            prefix = f"language_model.model.layers.{l}"

            # Stack text experts (0 to num_text_experts-1) into switch_mlp
            for m in ["gate_proj", "down_proj", "up_proj"]:
                for k in ["weight", "scales", "biases"]:
                    if f"{prefix}.mlp.experts.0.{m}.{k}" in weights:
                        to_join = [
                            weights.pop(f"{prefix}.mlp.experts.{e}.{m}.{k}")
                            for e in range(num_text_experts)
                        ]
                        weights[f"{prefix}.mlp.switch_mlp.{m}.{k}"] = mx.stack(to_join)

            # Stack multimodal experts (num_text_experts to num_text_experts+num_mm_experts-1) into switch_mlp_1
            if num_mm_experts > 0:
                for m in ["gate_proj", "down_proj", "up_proj"]:
                    for k in ["weight", "scales", "biases"]:
                        first_mm_expert = num_text_experts
                        if f"{prefix}.mlp.experts.{first_mm_expert}.{m}.{k}" in weights:
                            to_join = [
                                weights.pop(f"{prefix}.mlp.experts.{e}.{m}.{k}")
                                for e in range(
                                    num_text_experts, num_text_experts + num_mm_experts
                                )
                            ]
                            weights[f"{prefix}.mlp.switch_mlp_1.{m}.{k}"] = mx.stack(
                                to_join
                            )

            # Transpose gate weights if needed (HuggingFace uses [in, out], MLX uses [out, in])
            # MLX nn.Linear(in=2560, out=64) expects shape (64, 2560), HF provides (2560, 64)
            gate_key = f"{prefix}.mlp.gate.weight"
            if gate_key in weights:
                w = weights[gate_key]
                # Only transpose if shape is (hidden_size, num_experts) not (num_experts, hidden_size)
                if w.shape[0] > w.shape[1]:  # (2560, 64) needs transpose
                    weights[gate_key] = w.T

            # Rename gate.weight_1 to gate_1.weight for multimodal gate and transpose
            gate_1_key = f"{prefix}.mlp.gate.weight_1"
            if gate_1_key in weights:
                w = weights.pop(gate_1_key)
                if w.shape[0] > w.shape[1]:  # Only transpose if needed
                    w = w.T
                weights[f"{prefix}.mlp.gate_1.weight"] = w

            # Handle e_score_correction_bias
            # HuggingFace stores as [2, num_experts] - row 0 for text, row 1 for multimodal
            bias_key = f"{prefix}.mlp.moe_statics.e_score_correction_bias"
            if bias_key in weights:
                bias = weights.pop(bias_key)
                if bias.ndim == 2 and bias.shape[0] == 2:
                    # Split into text and multimodal biases
                    weights[f"{prefix}.mlp.e_score_correction_bias"] = bias[0]
                    if num_mm_experts > 0:
                        weights[f"{prefix}.mlp.e_score_correction_bias_1"] = bias[1]
                else:
                    # Single bias (squeeze if needed)
                    if bias.ndim > 1:
                        bias = bias.squeeze()
                    weights[f"{prefix}.mlp.e_score_correction_bias"] = bias

        # Remove lm_head if tie_word_embeddings is True
        if self.args.tie_word_embeddings:
            lm_head_key = "language_model.lm_head.weight"
            if lm_head_key in weights:
                weights.pop(lm_head_key)

        return weights



import re
from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from ..base import InputEmbeddingsFeatures
from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


def masked_scatter(
    final_embedding: mx.array,
    image_mask_expanded: mx.array,
    scaled_image_features: mx.array,
):
    # Reshape the tensors to 1D
    final_embedding_shape = final_embedding.shape
    scaled_image_features_flattened = mx.flatten(scaled_image_features)
    final_embedding_flattened = mx.flatten(final_embedding)
    image_mask_expanded_flattened = mx.flatten(image_mask_expanded)

    # Scatter the scaled image features into the special image token positions
    image_positions = mx.array(np.where(image_mask_expanded_flattened)[0], mx.uint32)
    final_embedding_flattened[image_positions] = scaled_image_features_flattened

    # Reshape back to the original shape
    final_embedding = mx.reshape(final_embedding_flattened, final_embedding_shape)

    return final_embedding


class MLP(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        input_size = config.vision_config.hidden_size * (config.scale_factor**2)
        output_size = config.text_config.hidden_size
        self.proj = nn.Linear(input_size, output_size, bias=False)

    def __call__(self, x):
        return self.proj(x)


class Idefics3Connector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.scale_factor = config.scale_factor
        self.modality_projection = MLP(config)

    def pixel_shuffle(self, x, scale_factor=2):
        bsz, seq, embed_dim = x.shape
        height = width = int(seq**0.5)
        x = x.reshape(bsz, height, width, embed_dim)
        x = x.reshape(bsz, height, int(width / scale_factor), embed_dim * scale_factor)
        x = x.transpose(0, 2, 1, 3)
        x = x.reshape(
            bsz,
            int(width / scale_factor),
            int(height / scale_factor),
            embed_dim * (scale_factor**2),
        )
        x = x.transpose(0, 2, 1, 3)
        x = x.reshape(bsz, int(seq / (scale_factor**2)), embed_dim * (scale_factor**2))
        return x

    def __call__(self, image_hidden_states):
        image_hidden_states = self.pixel_shuffle(image_hidden_states, self.scale_factor)
        image_hidden_states = self.modality_projection(image_hidden_states)
        return image_hidden_states


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.model_type = config.model_type
        self.config = config

        self.vision_model = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.connector = Idefics3Connector(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        pixel_attention_mask = kwargs.get("pixel_attention_mask", None)

        if pixel_values is None:
            return InputEmbeddingsFeatures(
                inputs_embeds=self.language_model.embed_tokens(input_ids)
            )

        inputs_embeds = self.language_model.embed_tokens(input_ids)

        batch_size, num_images, num_channels, height, width = pixel_values.shape
        pixel_values = pixel_values.reshape(
            batch_size * num_images, num_channels, height, width
        )

        # Remove padding images - padding image are full 0.
        nb_values_per_image = np.prod(pixel_values.shape[1:])
        real_images_mask = (pixel_values == 0.0).sum(
            axis=(-1, -2, -3)
        ) != nb_values_per_image
        real_images_inds = np.where(real_images_mask)[0].tolist()
        pixel_values = pixel_values[real_images_inds, ...]

        if pixel_attention_mask is None:
            pixel_attention_mask = mx.ones(
                (pixel_values.size(0), pixel_values.size(2), pixel_values.size(3)),
                dtype=mx.bool,
            )
        else:
            # Remove padding images from the mask
            pixel_attention_mask = pixel_attention_mask.reshape(
                batch_size * num_images, height, width
            )
            pixel_attention_mask = pixel_attention_mask[real_images_inds]

        patch_size = self.config.vision_config.patch_size
        batch_size, height, width = pixel_attention_mask.shape

        # Calculate number of patches
        patches_h = height // patch_size
        patches_w = width // patch_size

        # Reshape to extract patches
        reshaped = pixel_attention_mask[
            :, : patches_h * patch_size, : patches_w * patch_size
        ]
        reshaped = reshaped.reshape(
            batch_size, patches_h, patch_size, patches_w, patch_size
        )
        reshaped = reshaped.transpose(
            0, 1, 3, 2, 4
        )  # (batch, patches_h, patches_w, patch_size, patch_size)

        # Sum over patch dimensions and check if any pixels are active
        patch_attention_mask = reshaped.sum(axis=(-1, -2)) > 0

        pooler_output, *_ = self.vision_model(
            pixel_values.transpose(0, 2, 3, 1),
            patch_attention_mask=patch_attention_mask,
            output_hidden_states=True,
        )

        image_features = pooler_output.astype(pixel_values.dtype)
        image_features = self.connector(image_features)

        final_inputs_embeds = self._prepare_inputs_for_multimodal(
            image_features, inputs_embeds, input_ids
        )
        return InputEmbeddingsFeatures(inputs_embeds=final_inputs_embeds)

    def _prepare_inputs_for_multimodal(self, image_features, inputs_embeds, input_ids):
        special_image_mask = input_ids == self.config.image_token_index
        n_image_tokens = special_image_mask.sum()
        special_image_mask = special_image_mask[..., None]
        special_image_mask = mx.broadcast_to(special_image_mask, inputs_embeds.shape)

        n_image_features = image_features.shape[0]
        n_image_mask_elements = special_image_mask.sum()
        if n_image_mask_elements != image_features.size:
            raise ValueError(
                f"Image features and image tokens do not match: tokens: {n_image_tokens}, features {n_image_features}"
            )

        inputs_embeds = masked_scatter(
            inputs_embeds, special_image_mask, image_features
        )

        return inputs_embeds

    @property
    def layers(self):
        return self.language_model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        cache=None,
        **kwargs,
    ):
        input_embeddings_features = self.get_input_embeddings(
            input_ids, pixel_values, **kwargs
        )
        logits = self.language_model(
            inputs=input_ids,
            cache=cache,
            inputs_embeds=input_embeddings_features.inputs_embeds,
        )
        return logits

    def sanitize(self, weights):
        weights = {
            (
                f"{k.split('.', 1)[1]}"
                if re.match(r"^model\.", k)
                else (f"language_model.{k}" if re.match(r"^lm_head\.", k) else k)
            ): v
            for k, v in weights.items()
        }

        weights = {
            (
                f"language_model.{k.split('.', 1)[1]}"
                if re.match(
                    r"^text_model\.",
                    k,
                )
                else k
            ): v
            for k, v in weights.items()
        }

        return weights
